# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The chip class contains information about the chip and its conections. This information can be reached through the Platform class through a platform object by accessing platform.chip.

Each node in the chip can be a Coil, Coupler, Port, Qubit or Resonator. Each node also has an alias assigned to it. For example, the alias for Qubit nodes is typically ``qubit_n`` where n is the qubit index.

The qubit connectivity (chip topology) can be accessed by calling chip.get_topology(), which returns a networkx graph of the qubits, e.g.

.. code-block:: python

    import networkx as nx
    platform = ql.build_platform(runcard="runcard.yml")
    g = platform.chip.get_topology()
    nx.draw(g, with_labels=True)


.. image:: chip_images/chip_topo.png
   :scale: 60 %
   :alt: alternate text
   :align: center

.. currentmodule:: qililab.chip

Chip Class
~~~~~~~~~~


.. autosummary::
    :toctree: api

    ~Chip

Nodes
~~~~~

.. autosummary::
    :toctree: api

    ~Node
    ~Port
    ~Qubit
    ~Resonator
    ~Coupler
    ~Coil
"""
from .chip import Chip
from .node import Node
from .nodes import Coil, Coupler, Port, Qubit, Resonator



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Drivers for SpiRack and its corresponging channels: D5aDacChannel & S4gDacChannel."""
from typing import Union

from qblox_instruments import SpiRack as QcodesSpiRack
from qblox_instruments.qcodes_drivers.spi_rack_modules.d5a_module import D5aDacChannel as QcodesD5aDacChannel
from qblox_instruments.qcodes_drivers.spi_rack_modules.d5a_module import D5aModule as QcodesD5aModule
from qblox_instruments.qcodes_drivers.spi_rack_modules.s4g_module import S4gDacChannel as QcodesS4gDacChannel
from qblox_instruments.qcodes_drivers.spi_rack_modules.s4g_module import S4gModule as QcodesS4gModule
from qcodes import Instrument
from qcodes.instrument.channel import ChannelTuple, InstrumentModule

from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import BaseInstrument, CurrentSource, VoltageSource


# MAIN SpiRack CLASS
@InstrumentDriverFactory.register
class SpiRack(QcodesSpiRack, BaseInstrument):  # pylint: disable=abstract-method
    """
    Qililab's driver for the Qblox SpiRack.

    SPI rack driver class based on `QCoDeS <https://qcodes.github.io/Qcodes/>`_.

    This class is initialize as in Qblox but changing the references to qililab classes instead.

    Args:
        name (str): Instrument name.
        address (str): COM port used by SPI rack controller unit (e.g. "COM4")
    """

    def __init__(self, name: str, address: str, **kwargs):
        super().__init__(name, address, **kwargs)

        self._MODULES_MAP["S4g"] = S4gModule
        self._MODULES_MAP["D5a"] = D5aModule

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name


# MODULE CLASSES that select the channels
class D5aModule(QcodesD5aModule, BaseInstrument):
    """
    Qililab's driver for the Qblox D5a Module.

    `QCoDeS <https://qcodes.github.io/Qcodes/>`_ style instrument channel driver for the D5a SPI module.

    Args:
        parent(SpiRack): Reference to the :class:`~qblox_instruments.SpiRack` parent object. This is handled by
            the :func:`~qblox_instruments.SpiRack.add_spi_module` function.
        name (str): Name given to the InstrumentChannel.
        address (int): Module number set on the hardware.
    Raises: ValueError: Length of the dac names list does not match the number of dacs.
        is_dummy (bool, optional): _description_. Defaults to False.
    """

    def __init__(self, parent: Instrument, name: str, address: int, **kwargs):
        super().__init__(parent, name, address, **kwargs)

        self.submodules: dict[str, Union[InstrumentModule, ChannelTuple]] = {}  # resetting superclass submodules
        self.instrument_modules: dict[str, InstrumentModule] = {}  # resetting superclass instrument modules
        self._channel_lists: dict[str, ChannelTuple] = {}  # resetting superclass channel lists
        for dac, old_channel in enumerate(self._channels):
            new_channel = D5aDacChannel(self, old_channel._chan_name, dac)
            self._channels[dac] = new_channel
            self.add_submodule(old_channel._chan_name, new_channel)

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name


class S4gModule(QcodesS4gModule, BaseInstrument):
    """
    Qililab's driver for the Qblox S4g Module.

    `QCoDeS <https://qcodes.github.io/Qcodes/>`_ style instrument channel driver for the S4g SPI module.

    Args:
        parent(SpiRack): Reference to the :class:`~qblox_instruments.SpiRack` parent object. This is handled by
            the :func:`~qblox_instruments.SpiRack.add_spi_module` function.
        name (str): Name given to the InstrumentChannel.
        address (int): Module number set on the hardware.

    Raises:
        ValueError: Length of the dac names list does not match the number of dacs.
    """

    def __init__(self, parent: Instrument, name: str, address: int, **kwargs):
        super().__init__(parent, name, address, **kwargs)

        self.submodules: dict[str, Union[InstrumentModule, ChannelTuple]] = {}  # resetting superclass submodules
        self.instrument_modules: dict[str, InstrumentModule] = {}  # resetting superclass instrument modules
        self._channel_lists: dict[str, ChannelTuple] = {}  # resetting superclass channel lists
        for dac, old_channel in enumerate(self._channels):
            new_channel = S4gDacChannel(self, old_channel._chan_name, dac)
            self._channels[dac] = new_channel
            self.add_submodule(old_channel._chan_name, new_channel)

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name


# CHANNELS CLASSES that act as the corresponding Voltage/Current sources.
class D5aDacChannel(QcodesD5aDacChannel, VoltageSource):
    """
    Qililab's driver for the Qblox D5a DAC Channel x16, acting as VoltageSource.

    `QCoDeS <https://qcodes.github.io/Qcodes/>`_ style instrument channel driver for the dac channels of the D5a module.

    Args:
        parent (D5aModule): Reference to the parent :class:`~qblox_instruments.qcodes_drivers.spi_rack_modules.D5aModule`
        name (str): Name for the instrument channel
        dac (int): Number of the dac that this channel corresponds to
    """

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name

    def on(self) -> None:
        """Start D5aDacChannel"""

    def off(self) -> None:
        """Turn output off"""
        self.set(param_name="voltage", value=0)


class S4gDacChannel(QcodesS4gDacChannel, CurrentSource):
    """
    Qililab's driver for the Qblox S4g DAC Channel x4, acting as CurrentSource.

    `QCoDeS <https://qcodes.github.io/Qcodes/>`_ style instrument channel driver for the dac channels of the S4g module.

    Args:
        parent (S4gModule): Reference to the parent :class:`~qblox_instruments.qcodes_drivers.spi_rack_modules.S4gModule`
        name (str): Name for the instrument channel
        dac (int): Number of the dac that this channel corresponds to
    """

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name

    def on(self) -> None:
        """Start S4gDacChannel"""

    def off(self) -> None:
        """Turn output off"""
        self.set(param_name="current", value=0)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from qblox_instruments.qcodes_drivers.sequencer import Sequencer
from qpysequence.acquisitions import Acquisitions
from qpysequence.library import long_wait
from qpysequence.program import Block, Loop, Program, Register
from qpysequence.program.instructions import Play, ResetPh, SetAwgGain, SetPh, Stop
from qpysequence.sequence import Sequence as QpySequence
from qpysequence.utils.constants import AWG_MAX_GAIN
from qpysequence.waveforms import Waveforms
from qpysequence.weights import Weights

from qililab.config import logger
from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import AWG
from qililab.pulse import PulseBusSchedule, PulseShape


@InstrumentDriverFactory.register
class SequencerQCM(Sequencer, AWG):
    """Qililab's driver for QBlox-instruments Sequencer

    Args:
        parent (Instrument): Parent for the sequencer instance.
        name (str): Sequencer name
        seq_idx (int): sequencer identifier index
    """

    _MIN_WAIT_TIME: int = 4

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name

    def execute(self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int):
        """Execute a PulseBusSchedule on the instrument.

        Args:
            pulse_bus_schedule (PulseBusSchedule): PulseBusSchedule to be translated into QASM program and executed.
            nshots (int): number of shots / hardware average
            repetition_duration (int): repetition duration
            num_bins (int): number of bins
        """
        sequence = self._translate_pulse_bus_schedule(pulse_bus_schedule, nshots, repetition_duration, num_bins)
        self.set("sequence", sequence.todict())
        self.parent.arm_sequencer(sequencer=self.seq_idx)
        self.parent.start_sequencer(sequencer=self.seq_idx)

    def _translate_pulse_bus_schedule(
        self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int
    ):
        """Translate a pulse sequence into a Q1ASM program and a waveform dictionary.

        Args:
            pulse_bus_schedule (PulseBusSchedule): Pulse bus schedule to translate.
            nshots (int): number of shots / hardware average
            repetition_duration (int): repetition duration
            num_bins (int): number of bins

        Returns:
            Sequence: Qblox Sequence object containing the program and waveforms.
        """
        waveforms = self._generate_waveforms(pulse_bus_schedule=pulse_bus_schedule)
        acquisitions = self._generate_acquisitions(num_bins=num_bins)
        program = self._generate_program(
            pulse_bus_schedule=pulse_bus_schedule,
            waveforms=waveforms,
            nshots=nshots,
            repetition_duration=repetition_duration,
            num_bins=num_bins,
        )
        weights = self._generate_weights()

        return QpySequence(program=program, waveforms=waveforms, weights=weights, acquisitions=acquisitions)

    def _generate_weights(self) -> Weights:
        """Generate acquisition weights.

        Returns:
            dict: Acquisition weights.
        """
        return Weights()

    def _generate_acquisitions(self, num_bins: int) -> Acquisitions:  # pylint: disable=unused-argument
        """Generate Acquisitions object.

        Args:
            num_bins (int): number of bins

        Returns:
            Acquisitions: Acquisitions object.
        """
        return Acquisitions()

    def _generate_waveforms(self, pulse_bus_schedule: PulseBusSchedule):
        """Generate I and Q waveforms from a PulseSequence object.

        Args:
            pulse_bus_schedule (PulseBusSchedule): PulseSequence object.

        Returns:
            Waveforms: Waveforms object containing the generated waveforms.
        """
        waveforms = Waveforms()

        unique_pulses: list[tuple[int, PulseShape]] = []

        for pulse_event in pulse_bus_schedule.timeline:
            if (pulse_event.duration, pulse_event.pulse.pulse_shape) not in unique_pulses:
                unique_pulses.append((pulse_event.duration, pulse_event.pulse.pulse_shape))
                amp = pulse_event.pulse.amplitude
                sign = 1 if amp >= 0 else -1
                envelope = pulse_event.envelope(amplitude=sign * 1.0)
                real = np.real(envelope)
                imag = np.imag(envelope)
                pair = (real, imag)
                waveforms.add_pair(pair=pair, name=pulse_event.pulse.label())

        return waveforms

    def _init_weights_registers(self, registers: tuple[Register, Register], values: tuple[int, int], program: Program):
        """Initialize the weights `registers` to the `values` specified and place the required instructions in the
        setup block of the `program`."""

    def _append_acquire_instruction(
        self, loop: Loop, bin_index: Register | int, weight_regs: tuple[Register, Register]
    ):
        """Append an acquire instruction to the loop."""

    def _generate_program(  # pylint: disable=too-many-locals
        self,
        pulse_bus_schedule: PulseBusSchedule,
        waveforms: Waveforms,
        nshots: int,
        repetition_duration: int,
        num_bins: int,
    ):
        """Generate Q1ASM program

        Args:
            pulse_sequence (PulseSequence): pulse sequence
            waveforms (Waveforms): waveforms
            nshots (int): number of shots / hardware average
            repetition_duration (int): repetition duration
            num_bins (int): number of bins

        Returns:
            Program: Q1ASM program.
        """
        # Define program's blocks
        program = Program()
        # Create registers with 0 and 1 (necessary for qblox)
        weight_registers = Register(), Register()
        self._init_weights_registers(registers=weight_registers, values=(0, 1), program=program)
        avg_loop = Loop(name="average", begin=nshots)
        bin_loop = Loop(name="bin", begin=0, end=num_bins, step=1)
        avg_loop.append_component(bin_loop)
        program.append_block(avg_loop)
        stop = Block(name="stop")
        stop.append_component(Stop())
        program.append_block(block=stop)
        timeline = pulse_bus_schedule.timeline
        if len(timeline) > 0 and timeline[0].start_time != 0:
            bin_loop.append_component(long_wait(wait_time=int(timeline[0].start_time)))

        for i, pulse_event in enumerate(timeline):
            print("Pulse label: ", pulse_event.pulse.label())
            waveform_pair = waveforms.find_pair_by_name(pulse_event.pulse.label())
            wait_time = timeline[i + 1].start_time - pulse_event.start_time if (i < (len(timeline) - 1)) else 4
            bin_loop.append_component(ResetPh())
            gain = int(np.abs(pulse_event.pulse.amplitude) * AWG_MAX_GAIN)  # np.abs() needed for negative pulses
            bin_loop.append_component(SetAwgGain(gain_0=gain, gain_1=gain))
            phase = int((pulse_event.pulse.phase % (2 * np.pi)) * 1e9 / (2 * np.pi))
            bin_loop.append_component(SetPh(phase=phase))
            bin_loop.append_component(
                Play(
                    waveform_0=waveform_pair.waveform_i.index,
                    waveform_1=waveform_pair.waveform_q.index,
                    wait_time=int(wait_time),
                )
            )
        self._append_acquire_instruction(
            loop=bin_loop, bin_index=bin_loop.counter_register, weight_regs=weight_registers
        )
        if repetition_duration is not None:
            wait_time = repetition_duration - bin_loop.duration_iter
            if wait_time > self._MIN_WAIT_TIME:
                bin_loop.append_component(long_wait(wait_time=wait_time))

        logger.info("Q1ASM program: \n %s", repr(program))  # pylint: disable=protected-access

        return program



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" BaseExperiment class."""
import os
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from queue import Empty, Queue
from threading import Thread

import numpy as np
from qcodes.instrument import Instrument as QcodesInstrument
from ruamel.yaml import YAML
from tqdm.auto import tqdm

from qililab.chip import Node
from qililab.config import __version__, logger
from qililab.constants import DATA, EXPERIMENT, EXPERIMENT_FILENAME, RESULTS_FILENAME, RUNCARD
from qililab.execution import EXECUTION_BUILDER, ExecutionManager
from qililab.platform.platform import Platform
from qililab.result.results import Results
from qililab.settings import Runcard
from qililab.settings.gate_event_settings import GateEventSettings
from qililab.typings.enums import Instrument, Parameter
from qililab.typings.experiment import ExperimentOptions
from qililab.utils.live_plot import LivePlot
from qililab.utils.loop import Loop
from qililab.utils.util_loops import compute_shapes_from_loops


class BaseExperiment(ABC):
    """BaseExperiment class"""

    # Specify the types of the attributes that are not defined during initialization
    execution_manager: ExecutionManager
    results: Results
    results_path: Path | None
    _plot: LivePlot | None
    _remote_id: int

    def __init__(self, platform: Platform, options: ExperimentOptions = ExperimentOptions()):
        self.platform = platform
        self.options = options

    def build_execution(self):
        """Creates the ``ExecutionManager`` class from loops."""
        # Build ``ExecutionManager`` class
        self.execution_manager = EXECUTION_BUILDER.build_from_loops(platform=self.platform, loops=self.options.loops)

    def run(self, save_experiment=True, save_results=True) -> Results:
        """This method is responsible for:
        * Creating the live plotting (if connection is provided).
        * Preparing the `Results` class and the `results.yml` file.
        * Looping over all given loops and/or software averages. And for each loop:
            * Saving the results to the ``results.yml`` file.
            * Sending the data to the live plotting (if asked to).
            * Save the results to the ``results`` attribute.
            * Save the results to the remote database (if asked to).
        """
        # Generate live plotting
        if self.platform.connection is None:
            self._plot = None
        else:
            # TODO: Live plotting should be able to hable num_schedules=0
            self._plot = LivePlot(
                connection=self.platform.connection,
                loops=self.options.loops or [],
                num_schedules=1,
                title=self.options.name,
            )

        if not hasattr(self, "execution_manager"):
            raise ValueError("Please build the execution_manager before running an experiment.")

        # Prepares the results
        self.results, self.results_path = self.prepare_results(
            save_experiment=save_experiment, save_results=save_results
        )

        data_queue: Queue = Queue()  # queue used to store the experiment results
        self._asynchronous_data_handling(queue=data_queue)
        self._execute_recursive_loops(loops=self.options.loops, queue=data_queue)

        if self.options.remote_save:
            self.remote_save_experiment()

        return self.results

    def _asynchronous_data_handling(self, queue: Queue):
        """Starts a thread that asynchronously gets the results from the queue, sends them to the live plot (if any)
        and saves them to a file.

        If no items are received in the queue for 5 seconds, the thread will exit.

        Args:
            queue (Queue): Queue used to store the experiment results.
        """
        timeout = max(5, 10 * self.hardware_average * self.repetition_duration * self.num_bins * 1e-9)

        def _threaded_function():
            """Asynchronous thread."""
            while True:
                try:
                    result = queue.get(timeout=timeout)  # get new result from the queue
                except Empty:
                    return  # exit thread if no results are received for 10 times the duration of the program

                if self._plot is not None:
                    acq = result.acquisitions()
                    i = np.array(acq["i"])
                    q = np.array(acq["q"])
                    amplitude = 20 * np.log10(np.abs(i + 1j * q)).astype(np.float64)
                    self._plot.send_points(value=amplitude[0])

                if self.results_path is not None:
                    with open(file=self.results_path / "results.yml", mode="a", encoding="utf8") as data_file:
                        result_dict = result.to_dict()
                        YAML().dump(data=[result_dict], stream=data_file)

        thread = Thread(target=_threaded_function)
        thread.start()

    def execute(self, save_experiment=True, save_results=True) -> Results:
        """Runs the whole execution pipeline, which includes the following steps:

            * Connect to the instruments.
            * Apply settings of the runcard to the instruments.
            * Translate circuit into pulses and create the ``ExecutionManager`` class.
            * Turn on instruments.
            * Create the results files & class and connect to live plotting.
            * Runs the experiment.
            * Turn off instruments.
            * Disconnect from the instruments.
            * Return results.

        Returns:
            Results: execution results
        """
        self.platform.connect()
        self.platform.initial_setup()
        self.build_execution()
        self.platform.turn_on_instruments()
        results = self.run(save_experiment=save_experiment, save_results=save_results)
        self.platform.turn_off_instruments()
        self.platform.disconnect()
        QcodesInstrument.close_all()
        return results

    def remote_save_experiment(self) -> None:
        """Saves the experiment and the results to the remote database and updates the ``_remote_id`` attribute.

        Raises:
            ValueError: if connection is not specified
        """
        if self.platform.connection is None:
            return

        logger.debug("Sending experiment and results to remote database.")
        self._remote_id = self.platform.connection.save_experiment(
            name=self.options.name,
            description=self.options.description,
            experiment_dict=self.to_dict(),
            results_dict=self.results.to_dict(),
            device_id=self.platform.device_id,
            user_id=self.platform.connection.user_id,
            qililab_version=__version__,
            favorite=False,
        )

    @abstractmethod
    def _execute_recursive_loops(self, loops: list[Loop] | None, queue: Queue, depth=0):
        """Loop over all the values defined in the Loop class and change the parameters of the chosen instruments.

        Args:
            loops (list[Loop]): list of Loop classes containing the info of one or more Platform element and the
            parameter values to loop over.
            depth (int): depth of the recursive loop.
        """

    def _process_loops(self, loops: list[Loop], queue: Queue, depth: int):
        """Loop over the loop values, change the element's parameter and call the recursive_loop function.

        Args:
            loops (list[Loop]): list of Loop classes containing the info of one or more Platform element and the
            parameter values to loop over.
            depth (int): depth of the recursive loop.
        """
        is_the_top_loop = all(loop.previous is False for loop in loops)

        with tqdm(total=min(len(loop.values) for loop in loops), position=depth, leave=is_the_top_loop) as pbar:
            loop_ranges = [loop.values for loop in loops]

            for values in zip(*loop_ranges):
                self._update_tqdm_bar(loops=loops, values=values, pbar=pbar)
                filtered_loops, filtered_values = self._filter_loops_values_with_external_parameters(
                    values=values,
                    loops=loops,
                )
                self._update_parameters_from_loops(values=filtered_values, loops=filtered_loops)
                inner_loops = list(filter(None, [loop.loop for loop in loops]))
                self._execute_recursive_loops(loops=inner_loops, queue=queue, depth=depth + 1)

    def _update_tqdm_bar(self, loops: list[Loop], values: tuple[float], pbar):
        """Updates TQDM bar"""
        description = []
        for value, loop in zip(values, loops):
            parameter_text = (
                loop.alias if loop.parameter == Parameter.EXTERNAL and loop.alias is not None else loop.parameter.value
            )
            description.append(f"{parameter_text}: {value}")
        pbar.set_description(" | ".join(description))
        pbar.update()

    def _filter_loops_values_with_external_parameters(self, values: tuple[float, ...], loops: list[Loop]):
        """filter loops and values removing those with external parameters"""
        if len(values) != len(loops):
            raise ValueError(f"Values list length: {len(values)} differ from loops list length: {len(loops)}.")
        filtered_loops = loops.copy()
        filtered_values = list(values).copy()
        for idx, loop in enumerate(filtered_loops):
            if loop.parameter == Parameter.EXTERNAL:
                filtered_loops.pop(idx)
                filtered_values.pop(idx)

        return filtered_loops, filtered_values

    def _update_parameters_from_loops(self, values: list[float], loops: list[Loop]):
        """update parameters from loops"""
        elements = [self.platform.get_element(alias=loop.alias) for loop in loops]

        for value, loop, element in zip(values, loops, elements):
            self.set_parameter(
                element=element, alias=loop.alias, parameter=loop.parameter, value=value, channel_id=loop.channel_id
            )

    def set_parameter(
        self,
        parameter: Parameter,
        value: float | str | bool,
        alias: str,
        element: Runcard.GatesSettings | Node | Instrument | None = None,
        channel_id: int | None = None,
    ):
        """Set parameter of a platform element.

        Args:
            parameter (Parameter): name of the parameter to change
            value (float): new value
            alias (str): alias of the element that contains the given parameter
            channel_id (int | None): channel id
        """
        if element is None:
            self.platform.set_parameter(alias=alias, parameter=Parameter(parameter), value=value, channel_id=channel_id)
        elif isinstance(element, Runcard.GatesSettings):
            element.set_parameter(alias=alias, parameter=parameter, value=value, channel_id=channel_id)
            self.build_execution()
        elif isinstance(element, list):  # if element is a list of GateEventSettings
            if all(isinstance(element_event, GateEventSettings) for element_event in element):  # type: ignore
                self.platform.set_parameter(alias=alias, parameter=Parameter(parameter), value=value)
                self.build_execution()
        else:
            element.set_parameter(parameter=parameter, value=value, channel_id=channel_id)  # type: ignore
            if parameter == Parameter.DELAY:
                self.build_execution()

    def to_dict(self):
        """Convert BaseExperiment into a dictionary.

        Returns:
            dict: Dictionary representation of the BaseExperiment class.
        """
        return {
            RUNCARD.PLATFORM: self.platform.to_dict(),
            EXPERIMENT.OPTIONS: self.options.to_dict(),
        }

    @classmethod
    @abstractmethod
    def from_dict(cls, dictionary: dict):
        """Load BaseExperiment from dictionary.

        Args:
            dictionary (dict): Dictionary description of an BaseExperiment.
        """

    def __str__(self):
        """String representation of a base experiment."""
        return f"BaseExperiment {self.options.name}:\n" + f"{str(self.platform)}\n" + f"{str(self.options)}"

    @property
    def software_average(self):
        """BaseExperiment 'software_average' property.
        Returns:
            int: settings.software_average.
        """
        return self.options.settings.software_average

    @property
    def hardware_average(self):
        """BaseExperiment 'hardware_average' property.
        Returns:
            int: settings.hardware_average.
        """
        return self.options.settings.hardware_average

    @property
    def num_bins(self):
        """BaseExperiment `num_bins` property.
        Returns
            int: settings.num_bins.
        """
        return self.options.settings.num_bins

    @property
    def repetition_duration(self):
        """BaseExperiment 'repetition_duration' property.
        Returns:
            int: settings.repetition_duration.
        """
        return self.options.settings.repetition_duration

    def prepare_results(self, save_experiment=True, save_results=True) -> tuple[Results, Path | None]:
        """Creates the ``Results`` class, creates the ``results.yml`` file where the results will be saved, and dumps
        the experiment data into this file.

        Args:
            options (ExperimentOptions): options of the experiment
            num_schedules (int): number of schedules
            experiment_serialized (dict): dictionary representing the current experiment

        Returns:
            tuple[Results, Path]: a tuple containing the ``Results`` class and the path to the ``results.yml`` file
        """
        # Create the ``Results`` class
        results = Results(
            software_average=self.options.settings.software_average,
            num_schedules=self.execution_manager.num_schedules,
            loops=self.options.loops,
        )

        if save_results or save_experiment:
            # Create the folders & files needed to save the results locally
            results_path = self._path_to_results_folder()
            self._create_results_file(results_path)
            if save_experiment:
                # Dump the experiment data into the created file
                with open(file=results_path / EXPERIMENT_FILENAME, mode="w", encoding="utf-8") as experiment_file:
                    YAML().dump(data=self.to_dict(), stream=experiment_file)
        else:
            results_path = None

        return results, results_path

    def _path_to_results_folder(self) -> Path:
        """Creates a folder for the current day (if needed), creates another folder for the current timestamp
        (if needed) and returns the path to this last folder.

        Args:
            name (str): name to identify the folder besides the timestamp

        Returns:
            Path: Path to folder.
        """
        # Timestamp
        now = datetime.now()

        # Get path to DATA folder from environment
        daily_path = os.environ.get(DATA, None)
        if daily_path is None:
            raise ValueError("Environment variable DATA is not set.")

        # Generate path to the daily folder
        daily_path = Path(daily_path) / f"{now.year}{now.month:02d}{now.day:02d}"  # type: ignore

        # Check if folder exists, if not create one
        if not os.path.exists(daily_path):
            os.makedirs(daily_path)

        # Generate path to the results folder
        now_path = daily_path / f"{now.hour:02d}{now.minute:02d}{now.second:02d}_{self.options.name}"  # type: ignore

        # Check if folder exists, if not create one
        if not os.path.exists(now_path):
            os.makedirs(now_path)

        return now_path

    def _create_results_file(self, path: Path):
        """Create 'results.yml' file that contains all the information about the current experiment.

        Args:
            path (Path): Path to data folder.
        """

        data = {
            EXPERIMENT.SOFTWARE_AVERAGE: self.options.settings.software_average,
            EXPERIMENT.NUM_SCHEDULES: self.execution_manager.num_schedules,
            EXPERIMENT.SHAPE: [] if self.options.loops is None else compute_shapes_from_loops(loops=self.options.loops),
            EXPERIMENT.LOOPS: [loop.to_dict() for loop in self.options.loops]
            if self.options.loops is not None
            else None,
            EXPERIMENT.RESULTS: None,
        }
        with open(file=path / RESULTS_FILENAME, mode="w", encoding="utf-8") as results_file:
            YAML().dump(data=data, stream=results_file)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Experiment class."""
import copy
import itertools
from queue import Queue

from qibo.models.circuit import Circuit
from tqdm.auto import tqdm

from qililab.chip import Node
from qililab.circuit_transpiler import CircuitTranspiler
from qililab.config import __version__
from qililab.constants import EXPERIMENT, RUNCARD
from qililab.execution import EXECUTION_BUILDER
from qililab.experiment.base_experiment import BaseExperiment
from qililab.platform.platform import Platform
from qililab.pulse import PulseSchedule
from qililab.result.results import Results
from qililab.settings import Runcard
from qililab.typings.enums import Instrument, Parameter
from qililab.typings.experiment import ExperimentOptions
from qililab.utils.live_plot import LivePlot
from qililab.utils.loop import Loop


class Experiment(BaseExperiment):
    """Experiment class"""

    def __init__(
        self,
        platform: Platform,
        circuits: list[Circuit] | None = None,
        pulse_schedules: list[PulseSchedule] | None = None,
        options: ExperimentOptions = ExperimentOptions(),
    ):
        self.circuits = circuits or []
        self.pulse_schedules = pulse_schedules or []
        super().__init__(platform=platform, options=options)

    def build_execution(self):
        """Translates the list of circuits to pulse sequences (if needed) and creates the ``ExecutionManager`` class."""
        # Translate circuits into pulses if needed
        if self.circuits:
            transpiler = CircuitTranspiler(platform=self.platform)
            self.pulse_schedules = transpiler.circuit_to_pulses(circuits=self.circuits)
        # Build ``ExecutionManager`` class
        self.execution_manager = EXECUTION_BUILDER.build(platform=self.platform, pulse_schedules=self.pulse_schedules)

    def run(self, save_experiment=True, save_results=True) -> Results:
        """This method is responsible for:

        * Creating the live plotting (if connection is provided).

        * Preparing the `Results` class and the `results.yml` file.

        * Looping over all the given circuits, loops and/or software averages. And for each loop:

            * Generating and uploading the program corresponding to the circuit.

            * Executing the circuit.

            * Saving the results to the ``results.yml`` file.

            * Sending the data to the live plotting (if asked to).

            * Save the results to the ``results`` attribute.

            * Save the results to the remote database (if asked to).
        """
        # Generate live plotting
        if self.platform.connection is None:
            self._plot = None
        else:
            self._plot = LivePlot(
                connection=self.platform.connection,
                loops=self.options.loops or [],
                num_schedules=len(self.pulse_schedules),
                title=self.options.name,
            )

        if not hasattr(self, "execution_manager"):
            raise ValueError("Please build the execution_manager before running an experiment.")
        # Prepares the results
        self.results, self.results_path = self.prepare_results(
            save_experiment=save_experiment, save_results=save_results
        )
        num_schedules = self.execution_manager.num_schedules

        data_queue: Queue = Queue()  # queue used to store the experiment results
        self._asynchronous_data_handling(queue=data_queue)

        for idx, _ in itertools.product(
            tqdm(range(num_schedules), desc="Sequences", leave=False, disable=num_schedules == 1),
            range(self.software_average),
        ):
            self._execute_recursive_loops(loops=self.options.loops, idx=idx, queue=data_queue)

        if self.options.remote_save:
            self.remote_save_experiment()

        return self.results

    def _execute_recursive_loops(self, loops: list[Loop] | None, queue: Queue, depth=0, **kwargs):
        """Loop over all the values defined in the Loop class and change the parameters of the chosen instruments.

        Args:
            loops (list[Loop]): list of Loop classes containing the info of one or more Platform element and the
            parameter values to loop over.
            idx (int): index of the circuit to execute
            depth (int): depth of the recursive loop.
            kwargs (dict): optional paramters, valid optional parameters:
                idx (int): index of the circuit to execute
        """
        if "idx" not in kwargs:
            raise ValueError("Parameter 'idx' must be specified")
        idx = copy.deepcopy(kwargs["idx"])

        if loops is None or len(loops) == 0:
            result = self.platform.execute(
                program=self.pulse_schedules[idx],
                num_avg=self.hardware_average,
                repetition_duration=self.repetition_duration,
                num_bins=self.num_bins,
                queue=queue,
            )
            if result is not None:
                self.results.add(result)
            return

        self._process_loops(loops=loops, idx=idx, queue=queue, depth=depth)

    def _process_loops(self, loops: list[Loop], queue: Queue, depth: int, **kwargs):  # pylint: disable=unused-argument
        """Loop over the loop values, change the element's parameter and call the recursive_loop function.

        Args:
            loops (list[Loop]): list of Loop classes containing the info of one or more Platform element and the
            parameter values to loop over.
            depth (int): depth of the recursive loop.
            kwargs (dict): optional paramters, valid optional parameters:
                idx (int): index of the circuit to execute
        """

        if "idx" not in kwargs:
            raise ValueError("Parameter 'idx' must be specified")
        idx = copy.deepcopy(kwargs["idx"])

        is_the_top_loop = all(loop.previous is False for loop in loops)

        with tqdm(total=min(len(loop.values) for loop in loops), position=depth, leave=is_the_top_loop) as pbar:
            loop_ranges = [loop.values for loop in loops]

            for values in zip(*loop_ranges):
                self._update_tqdm_bar(loops=loops, values=values, pbar=pbar)
                filtered_loops, filtered_values = self._filter_loops_values_with_external_parameters(
                    values=values,
                    loops=loops,
                )
                self._update_parameters_from_loops(values=filtered_values, loops=filtered_loops)
                inner_loops = list(filter(None, [loop.loop for loop in loops]))
                self._execute_recursive_loops(idx=idx, loops=inner_loops, queue=queue, depth=depth + 1)

    def compile(self) -> list[dict]:
        """Returns a dictionary containing the compiled programs of each bus for each circuit / pulse schedule of the
        experiment.

        Returns:
            list[dict]: List of dictionaries, where each dictionary has a bus alias as keys and a list of
                compiled sequences as values.
        """
        if not hasattr(self, "execution_manager"):
            raise ValueError("Please build the execution_manager before compilation.")
        return [
            self.platform.compile(schedule, self.hardware_average, self.repetition_duration, self.num_bins)
            for schedule in self.pulse_schedules
        ]

    def draw(
        self,
        real: bool = True,
        imag: bool = True,
        absolute: bool = False,
        modulation: bool = True,
        linestyle: str = "-",
        resolution: float = 1.0,
        idx: int = 0,
    ):
        """Return figure with the waveforms/envelopes sent to each bus.

        You can plot any combination of the real (blue), imaginary (orange) and absolute (green) parts of the function.

        Args:
            real (bool): True to plot the real part of the function, False otherwise. Default to True.
            imag (bool): True to plot the imaginary part of the function, False otherwise. Default to True.
            absolute (bool): True to plot the absolute of the function, False otherwise. Default to False.
            modulation (bool): True to plot the modulated wave form, False for only envelope. Default to True.
            linestyle (str): lineplot ("-", "--", ":"), point plot (".", "o", "x") or any other linestyle matplotlib accepts. Defaults to "-".
            resolution (float, optional): The resolution of the pulses in ns. Defaults to 1.0.

        Returns:
            Figure: Matplotlib figure with the waveforms sent to each bus.
        """
        if not hasattr(self, "execution_manager"):
            raise ValueError("Please build the execution_manager before drawing the experiment.")

        return self.execution_manager.draw(
            real=real,
            imag=imag,
            absolute=absolute,
            modulation=modulation,
            linestyle=linestyle,
            resolution=resolution,
            idx=idx,
        )

    def to_dict(self):
        """Convert Experiment into a dictionary.

        Returns:
            dict: Dictionary representation of the Experiment class.
        """
        exp_dict = super().to_dict()
        exp_dict[EXPERIMENT.CIRCUITS] = [circuit.to_qasm() for circuit in self.circuits]
        exp_dict[EXPERIMENT.PULSE_SCHEDULES] = [pulse_schedule.to_dict() for pulse_schedule in self.pulse_schedules]
        return exp_dict

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Load Experiment from dictionary.

        Args:
            dictionary (dict): Dictionary description of a Experiment.
        """

        platform = Platform(runcard=Runcard(**dictionary[RUNCARD.PLATFORM]))
        circuits = (
            [Circuit.from_qasm(settings) for settings in dictionary[EXPERIMENT.CIRCUITS]]
            if EXPERIMENT.CIRCUITS in dictionary
            else []
        )
        pulse_schedules = (
            [PulseSchedule.from_dict(settings) for settings in dictionary[EXPERIMENT.PULSE_SCHEDULES]]
            if EXPERIMENT.PULSE_SCHEDULES in dictionary
            else []
        )
        experiment_options = ExperimentOptions.from_dict(dictionary[EXPERIMENT.OPTIONS])
        return Experiment(
            platform=platform,
            circuits=circuits,
            pulse_schedules=pulse_schedules,
            options=experiment_options,
        )

    def set_parameter(  # pylint: disable=too-many-arguments
        self,
        parameter: Parameter,
        value: float | str | bool,
        alias: str,
        element: Runcard.GatesSettings | Node | Instrument | None = None,
        channel_id: int | None = None,
    ):
        """Set parameter of a platform element.

        Args:
            parameter (Parameter): name of the parameter to change
            value (float): new value
            alias (str): alias of the element that contains the given parameter
            channel_id (int | None): channel id
        """
        if parameter == Parameter.GATE_PARAMETER:
            for circuit in self.circuits:
                parameters = list(sum(circuit.get_parameters(), ()))
                parameters[int(alias)] = value
                circuit.set_parameters(parameters)
            self.build_execution()
            return
        super().set_parameter(parameter=parameter, value=value, alias=alias, element=element, channel_id=channel_id)

    def __str__(self):
        """String representation of a Experiment."""
        base_exp_str = super().__str__()
        return f"{base_exp_str}\n{str(self.circuits)}\n{str(self.pulse_schedules)}\n"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Drivers for
- :class:`qcodes_contrib_drivers.drivers.ERAInstruments.ERASynth <qcodes_contrib_drivers.drivers.ERAInstruments.erasynth.ERASynth>`,
- :class:`qcodes_contrib_drivers.drivers.ERAInstruments.ERASynthPlus <qcodes_contrib_drivers.drivers.ERAInstruments.erasynth.ERASynthPlus>`, and
- :class:`qcodes_contrib_drivers.drivers.ERAInstruments.ERASynthPlusPlus <qcodes_contrib_drivers.drivers.ERAInstruments.erasynth.ERASynthPlusPlus>`
from
https://github.com/QCoDeS/Qcodes_contrib_drivers/blob/main/qcodes_contrib_drivers/drivers/ERAInstruments/erasynth.py

All 3 classes are the same but kept separated for clarity at higher layers of the code
We only have the ERASynthPlus in the lab so for the time being we don't need to define the others.
"""
from typing import Any

from qcodes.instrument import DelegateParameter
from qcodes_contrib_drivers.drivers.ERAInstruments import ERASynthPlus as QcdERASynthPlus

from qililab.drivers import parameters
from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import LocalOscillator


@InstrumentDriverFactory.register
class ERASynthPlus(QcdERASynthPlus, LocalOscillator):
    """Qililab's driver for the ERASynthPlus local oscillator

    QcdEraSynth: QCoDeS contributors driver for the ERASynthPlus instrument
    LocalOscillator: Qililab's local oscillator interface
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address, **kwargs)

        # change the name for frequency
        self.add_parameter(
            parameters.lo.frequency,
            label="Delegated parameter for local oscillator frequency",
            source=self.parameters["frequency"],
            parameter_class=DelegateParameter,
        )

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name



"""Tests for the Loop class."""
import numpy as np
import pytest

from qililab.typings import Parameter
from qililab.utils import Loop


@pytest.fixture(name="loop")
def fixture_loop() -> Loop:
    """Return Loop object"""
    return Loop(alias="X", parameter=Parameter.AMPLITUDE, values=np.linspace(0, 10, 10))


@pytest.fixture(name="nested_loop")
def fixture_nested_loop() -> Loop:
    """Return Loop object with a loop inside aka nested loop"""
    nested_loop = Loop(alias="X", parameter=Parameter.AMPLITUDE, values=np.linspace(0, 10, 10))
    nested_loop.loop = Loop(alias="Y", parameter=Parameter.AMPLITUDE, values=np.geomspace(2, 32, 5))
    return nested_loop


class TestLoop:
    """Unit tests checking the Loop attributes and methods"""

    def test_num_loops_property(self, loop: Loop):
        """Test num_loops property."""
        assert loop.num_loops == 1

    def test_num_loops_property_nested_loop(self, nested_loop: Loop):
        """Test num_loops property for a nested loop."""
        assert nested_loop.num_loops == 2

    def test_ranges_property(self, loop: Loop):
        """Terst the ranges property"""
        expected = np.array([np.linspace(0, 10, 10)], dtype=object)
        np.testing.assert_array_equal(loop.all_values, expected)

    def test_ranges_property_nested_loop(self, nested_loop: Loop):
        """Terst the ranges property for a nested loop"""
        range_1 = np.linspace(0, 10, 10)
        range_2 = np.geomspace(2, 32, 5)
        expected = np.array([range_1, range_2], dtype=object)
        for x, y in zip(nested_loop.all_values, expected):
            np.testing.assert_array_equal(x, y)

    def test_shapes_property(self, loop: Loop):
        """Test the shapes property"""
        assert loop.shape == [10]

    def test_shapes_property_nested_loop(self, nested_loop: Loop):
        """Test the shapes property for a nested loop"""
        assert nested_loop.shape == [10, 5]

    def test_loops_property(self, loop: Loop):
        """Test the loops property"""
        assert all(isinstance(x, Loop) for x in loop.loops)
        assert len(loop.loops) == 1

    def test_loops_property_nested_loop(self, nested_loop: Loop):
        """Test the loops property for a nested loop"""
        assert all(isinstance(x, Loop) for x in nested_loop.loops)
        assert len(nested_loop.loops) == 2

    def test_start_property(self, loop: Loop):
        """ "Test the start property"""
        assert loop.start == 0.0

    def test_stop_property(self, loop: Loop):
        """ "Test the stop property"""
        assert loop.stop == 10.0

    def test_num_property(self, loop: Loop):
        """ "Test the num property"""
        assert loop.num == 10



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qubit class"""
from dataclasses import dataclass

from qililab.chip.node import Node
from qililab.typings import NodeName
from qililab.utils import Factory


@Factory.register
@dataclass
class Qubit(Node):
    """This class is used to represent each of the qubits in a chip.

    Each qubit has a frequency associated to it and an index within the chip.

    Args:
        frequency (float): frequency of the qubit
        qubit_index (int): qubit index
    """

    name = NodeName.QUBIT
    frequency: float  #: frequency of the qubit
    qubit_index: int  #: index of the qubit



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Driver for the Qblox Pulsar class."""
from qblox_instruments.qcodes_drivers import Pulsar as QcodesPulsar
from qcodes.instrument.channel import ChannelTuple, InstrumentModule

from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces.base_instrument import BaseInstrument

from .sequencer_qcm import SequencerQCM
from .sequencer_qrm import SequencerQRM


@InstrumentDriverFactory.register
class Pulsar(QcodesPulsar, BaseInstrument):  # pylint: disable=abstract-method
    """Qililab's driver for QBlox-instruments Pulsar

    Args:
        name (str): Sequencer name
        address (str): Instrument address
    """

    def __init__(self, name: str, address: str | None = None, **kwargs):
        super().__init__(name, identifier=address, **kwargs)

        # Add sequencers
        self.submodules: dict[str, SequencerQCM | SequencerQRM] = {}  # resetting superclass submodules
        self.instrument_modules: dict[str, InstrumentModule] = {}  # resetting superclass instrument modules
        self._channel_lists: dict[str, ChannelTuple] = {}  # resetting superclass channel lists

        sequencer_class = SequencerQCM if self.is_qcm_type else SequencerQRM
        for seq_idx in range(6):
            seq = sequencer_class(parent=self, name=f"sequencer{seq_idx}", seq_idx=seq_idx)  # type: ignore
            self.add_submodule(f"sequencer{seq_idx}", seq)

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name



"""Class YokogawaGS200"""
from qcodes.instrument_drivers.yokogawa.GS200 import GS200 as Driver_GS200

from qililab.typings.instruments.device import Device


class YokogawaGS200(Driver_GS200, Device):
    """Typing class of the GS200 class defined by Qcodes."""



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" AWG with Digital To Analog Conversion (DAC) capabilities."""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Sequence, cast

from qililab.instruments.awg import AWG
from qililab.instruments.awg_settings.awg_adc_sequencer import AWGADCSequencer
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.result.result import Result
from qililab.typings.enums import AcquireTriggerMode, IntegrationMode, Parameter


class AWGAnalogDigitalConverter(AWG):
    """AWG with Digital To Analog Conversion (ADC) capabilities."""

    @dataclass
    class AWGAnalogDigitalConverterSettings(AWG.AWGSettings):
        """Contains the settings of a specific pulsar.

        Args:
            acquisition_delay_time (str): Time specified before starting the acquisition
            awg_sequencers (Sequence[AWGADCSequencer]): Properties of each AWG ADC sequencer
        """

        acquisition_delay_time: int  # ns
        awg_sequencers: Sequence[AWGADCSequencer]

    settings: AWGAnalogDigitalConverterSettings

    @abstractmethod
    def acquire_result(self) -> Result:
        """Read the result from the AWG instrument

        Returns:
            Result: Acquired result
        """

    def setup(  # pylint: disable=too-many-return-statements, too-many-branches
        self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None
    ):
        """set a specific parameter to the instrument"""
        if channel_id is None:
            if self.num_sequencers == 1:
                channel_id = 0
            else:
                raise ValueError("channel not specified to update instrument")
        if parameter == Parameter.ACQUISITION_DELAY_TIME:
            self._set_acquisition_delay_time(value=value)
            return
        if parameter == Parameter.SCOPE_HARDWARE_AVERAGING:
            self._set_scope_hardware_averaging(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.HARDWARE_DEMODULATION:
            self._set_hardware_demodulation(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SCOPE_ACQUIRE_TRIGGER_MODE:
            self._set_acquisition_mode(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.INTEGRATION_LENGTH:
            self._set_integration_length(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SAMPLING_RATE:
            self._set_sampling_rate(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.INTEGRATION_MODE:
            self._set_integration_mode(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SEQUENCE_TIMEOUT:
            self._set_sequence_timeout(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.ACQUISITION_TIMEOUT:
            self._set_acquisition_timeout(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SCOPE_STORE_ENABLED:
            self._set_scope_store_enabled(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.THRESHOLD:
            self._set_threshold(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.THRESHOLD_ROTATION:
            self._set_threshold_rotation(value=value, sequencer_id=channel_id)
            return

        raise ParameterNotFound(f"Invalid Parameter: {parameter.value}")

    @abstractmethod
    def _set_device_scope_hardware_averaging(self, value: bool, sequencer_id: int):
        """set scope_hardware_averaging for the specific channel

        Args:
            value (bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """

    @abstractmethod
    def _set_device_threshold(self, value: float, sequencer_id: int):
        """Set threshold value for the specific channel.

        Args:
            value (float): the threshold value
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float"""

    @abstractmethod
    def _set_device_threshold_rotation(self, value: float, sequencer_id: int):
        """Set threshold rotation value for the specific channel.

        Args:
            value (float): the threshold rotation value
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float"""

    @abstractmethod
    def _set_device_hardware_demodulation(self, value: bool, sequencer_id: int):
        """set hardware demodulation

        Args:
            value (bool): value to update
            sequencer_id (int): sequencer to update the value
        """

    @abstractmethod
    def _set_device_acquisition_mode(self, mode: AcquireTriggerMode, sequencer_id: int):
        """set acquisition_mode for the specific channel

        Args:
            mode (AcquireTriggerMode): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not string
        """

    @abstractmethod
    def _set_device_integration_length(self, value: int, sequencer_id: int):
        """set integration_length for the specific channel

        Args:
            value (int): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """

    @Instrument.CheckParameterValueBool
    def _set_scope_hardware_averaging(self, value: float | str | bool, sequencer_id: int):
        """set scope_hardware_averaging for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).scope_hardware_averaging = bool(value)

        if self.is_device_active():
            self._set_device_scope_hardware_averaging(value=bool(value), sequencer_id=sequencer_id)

    @Instrument.CheckParameterValueFloatOrInt
    def _set_threshold(self, value: float | str | bool, sequencer_id: int):
        """Set threshold value for the specific channel.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).threshold = float(value)

        if self.is_device_active():
            self._set_device_threshold(value=float(value), sequencer_id=sequencer_id)

    @Instrument.CheckParameterValueFloatOrInt
    def _set_threshold_rotation(self, value: float | str | bool, sequencer_id: int):
        """Set threshold rotation value for the specific channel.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).threshold_rotation = float(value)
        if self.is_device_active():
            self._set_device_threshold_rotation(value=float(value), sequencer_id=sequencer_id)

    @Instrument.CheckParameterValueBool
    def _set_hardware_demodulation(self, value: float | str | bool, sequencer_id: int):
        """set hardware demodulation

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).hardware_demodulation = bool(value)
        if self.is_device_active():
            self._set_device_hardware_demodulation(value=bool(value), sequencer_id=sequencer_id)

    def _set_acquisition_mode(self, value: float | str | bool | AcquireTriggerMode, sequencer_id: int):
        """set acquisition_mode for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not string
        """
        if not isinstance(value, AcquireTriggerMode) and not isinstance(value, str):
            raise ValueError(f"value must be a string or AcquireTriggerMode. Current type: {type(value)}")
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).scope_acquire_trigger_mode = AcquireTriggerMode(value)
        if self.is_device_active():
            self._set_device_acquisition_mode(mode=AcquireTriggerMode(value), sequencer_id=sequencer_id)

    @Instrument.CheckParameterValueFloatOrInt
    def _set_integration_length(self, value: int | float | str | bool, sequencer_id: int):
        """set integration_length for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).integration_length = int(value)
        if self.is_device_active():
            self._set_device_integration_length(value=int(value), sequencer_id=sequencer_id)

    @Instrument.CheckParameterValueFloatOrInt
    def _set_sampling_rate(self, value: int | float | str | bool, sequencer_id: int):
        """set sampling_rate for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).sampling_rate = float(value)

    def _set_integration_mode(self, value: float | str | bool | IntegrationMode, sequencer_id: int):
        """set integration_mode for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not string
        """
        if isinstance(value, (IntegrationMode, str)):
            cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).integration_mode = IntegrationMode(value)
        else:
            raise ValueError(f"value must be a string or IntegrationMode. Current type: {type(value)}")

    @Instrument.CheckParameterValueFloatOrInt
    def _set_sequence_timeout(self, value: int | float | str | bool, sequencer_id: int):
        """set sequence_timeout for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float or int
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).sequence_timeout = int(value)

    @Instrument.CheckParameterValueFloatOrInt
    def _set_acquisition_timeout(self, value: int | float | str | bool, sequencer_id: int):
        """set acquisition_timeout for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float or int
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).acquisition_timeout = int(value)

    @Instrument.CheckParameterValueFloatOrInt
    def _set_acquisition_delay_time(self, value: int | float | str | bool):
        """set acquisition_delaty_time for the specific channel

        Args:
            value (float | str | bool): value to update

        Raises:
            ValueError: when value type is not float or int
        """
        self.settings.acquisition_delay_time = int(value)

    @Instrument.CheckParameterValueBool
    def _set_scope_store_enabled(self, value: float | str | bool, sequencer_id: int):
        """set scope_store_enable

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        cast(AWGADCSequencer, self.get_sequencer(sequencer_id)).scope_store_enabled = bool(value)

    @property
    def acquisition_delay_time(self):
        """AWG 'delay_before_readout' property.
        Returns:
            int: settings.delay_before_readout.
        """
        return self.settings.acquisition_delay_time



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Port class."""
from dataclasses import dataclass

from qililab.chip.node import Node
from qililab.typings import NodeName
from qililab.typings.enums import Line
from qililab.utils import Factory


@Factory.register
@dataclass
class Port(Node):
    """This class is used to represent a port connected to the chip.

    Each port has a line associated to communicate with the chip. The different types of lines supported
    are flux, drive, feedline input and feedline output lines.

    Args:
        name (str): Name for the port
        line (Line): The type of line associated with the port
    """

    name = NodeName.PORT  #: Name for the port
    line: Line  #: The type of line associated with the port



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Node class"""
from dataclasses import dataclass

from qililab.settings import Settings
from qililab.typings import FactoryElement


@dataclass(kw_only=True)
class Node(Settings, FactoryElement):
    """This class is used to represent a node of the chip's graph.

    Each node is an element of the chip.

    Args:
        alias (str): Alias of the node
        nodes (list[str]): List of nodes within the node
    """

    alias: str  #: Alias of the node
    nodes: list[str]  #: List of nodes within the node

    def __str__(self):
        """String representation of a node."""
        return f"{self.alias}"



from collections import deque
from dataclasses import fields, is_dataclass
from enum import Enum, EnumMeta
from typing import Any, Protocol, Type, TypedDict, TypeVar, _ProtocolMeta, cast, runtime_checkable
from uuid import UUID

import numpy as np

T = TypeVar("T", bound="DictSerializable")


class DictSerializableObject(TypedDict):
    """Typed dictionary describing a serialized object."""

    type: str
    attributes: dict[str, Any]


def is_dict_serializable_object(obj: Any) -> bool:
    """Check if the object follows the schema of `DictSerializableObject` typed dictionary.

    Args:
        obj (Any): The object to check

    Returns:
        bool: True, if object follows the schema of `DictSerializableObject` typed dictionary.
    """
    if not isinstance(obj, dict):
        return False

    if "type" not in obj or not isinstance(obj["type"], str):
        return False

    if "attributes" not in obj or not isinstance(obj["attributes"], dict):
        return False

    return all(isinstance(key, str) for key in obj["attributes"])


class DictSerializableEnumMeta(EnumMeta):
    """Metaclass to be used in `DictSerializableEnum`. Automatically registers the enums to `DictSerializableFactory`."""

    def __new__(mcs, name, bases, namespace):  # pylint: disable=arguments-differ
        new_class = super().__new__(mcs, name, bases, namespace)
        if bases != (Enum,):  # Avoid registering the base Enum class
            DictSerializableFactory.register(name, new_class)
        return new_class


class DictSerializableEnum(Enum, metaclass=DictSerializableEnumMeta):
    """Class that extends Enum and implements the DictSerializable protocol.

    Methods:
        to_dict: Converts the enum into a dictionary format.
        from_dict: Creates an instance of the enum from a dictionary.
    """

    def to_dict(self) -> DictSerializableObject:
        """Serialize the enum member to a dictionary."""
        return {"type": self.__class__.__name__, "attributes": {"value": self.name}}

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "DictSerializableEnum":
        """Creates an instance of the enum from a dictionary."""
        return cls[data["value"]]


class DictSerializableFactory:
    """Factory to store information of `DictSerializable` classes."""

    _registry: dict[str, Type["DictSerializable"]] = {}

    @classmethod
    def register(cls, name: str, dict_serializable_class: Type["DictSerializable"]) -> None:
        """Register class."""
        cls._registry[name] = dict_serializable_class

    @classmethod
    def get(cls, name: str) -> Type["DictSerializable"]:
        """Get class by name."""
        if name not in cls._registry:
            raise ValueError(f"'{name}' is not a class that can be (de)serialized (from)to dictionary.")
        return cls._registry[name]


class DictSerializableMeta(_ProtocolMeta):
    """Metaclass to be used in `DictSerializable` protocol. Automatically registers any class inheriting from `DictSerializable` to `DictSerializableFactory`."""

    def __new__(  # pylint: disable=arguments-differ
        mcs: Type["DictSerializableMeta"], name: str, bases: tuple, namespace: dict[str, Any]
    ) -> Type["DictSerializable"]:
        new_class = super().__new__(mcs, name, bases, namespace)
        new_class = cast(Type["DictSerializable"], new_class)
        if bases != (Protocol,):  # Avoid registering the base DictSerializable protocol
            DictSerializableFactory.register(name, new_class)
        return new_class


@runtime_checkable
class DictSerializable(Protocol, metaclass=DictSerializableMeta):
    """A protocol that defines a serializable object which can be converted to and from a dictionary.

    This protocol is used to ensure that objects conforming to it can be serialized into a dictionary
    format and deserialized back into an object. It is particularly useful for objects that need to be
    easily converted to a format that can be JSON-serialized.

    Methods:
        to_dict: Converts the object into a dictionary format.
        from_dict: Creates an instance of the object from a dictionary.
    """

    def to_dict(self) -> DictSerializableObject:
        """Converts the object into a dictionary format.

        Returns:
            DictSerializableObject: A typed dictionary representing the serialized state of the object.
        """

        def process_element(element):  # pylint: disable=too-many-return-statements
            if isinstance(element, UUID):
                return {"type": UUID.__name__, "uuid": str(element)}
            if isinstance(element, deque):
                return {"type": deque.__name__, "elements": [process_element(item) for item in element]}
            if isinstance(element, list):
                return {"type": list.__name__, "elements": [process_element(item) for item in element]}
            if isinstance(element, tuple):
                return {"type": tuple.__name__, "elements": [process_element(item) for item in element]}
            if isinstance(element, set):
                return {"type": set.__name__, "elements": [process_element(item) for item in element]}
            if isinstance(element, np.ndarray):
                return {"type": np.ndarray.__name__, "elements": [process_element(item) for item in element.tolist()]}
            if isinstance(element, dict):
                return {
                    "type": dict.__name__,
                    "keys": list(element),
                    "elements": [process_element(element[key]) for key in element],
                }
            if isinstance(element, DictSerializable):
                return element.to_dict()
            return element

        attributes = {k: process_element(v) for k, v in vars(self).items()}

        return {"type": self.__class__.__name__, "attributes": attributes}

    @classmethod
    def from_dict(cls: Type[T], data: dict[str, Any]) -> T:
        """Creates an instance of the object from a dictionary.

        Args:
            data (dict[str, Any]): A dictionary containing the serialized state of the object.

        Returns:
            T: An instance of the object populated with data from the dictionary.
        """

        def process_attribute(attribute):  # pylint: disable=too-many-return-statements
            if isinstance(attribute, dict) and "type" in attribute and attribute["type"] == UUID.__name__:
                return UUID(attribute["uuid"])
            if isinstance(attribute, dict) and "type" in attribute and attribute["type"] == deque.__name__:
                return deque(process_attribute(item) for item in attribute["elements"])
            if isinstance(attribute, dict) and "type" in attribute and attribute["type"] == list.__name__:
                return list(process_attribute(item) for item in attribute["elements"])
            if isinstance(attribute, dict) and "type" in attribute and attribute["type"] == tuple.__name__:
                return tuple(process_attribute(item) for item in attribute["elements"])
            if isinstance(attribute, dict) and "type" in attribute and attribute["type"] == set.__name__:
                return set(process_attribute(item) for item in attribute["elements"])
            if isinstance(attribute, dict) and "type" in attribute and attribute["type"] == np.ndarray.__name__:
                return np.array([process_attribute(item) for item in attribute["elements"]])
            if isinstance(attribute, dict) and "type" in attribute and attribute["type"] == dict.__name__:
                return {key: process_attribute(item) for key, item in zip(attribute["keys"], attribute["elements"])}
            if is_dict_serializable_object(attribute):
                return from_dict(attribute)
            return attribute

        if is_dataclass(cls):
            # Create an instance with only the init-able fields
            init_fields = {f.name: process_attribute(data[f.name]) for f in fields(cls) if f.init}
            dataclass_instance = cls(**init_fields)

            # Check if the dataclass is frozen
            is_frozen = getattr(cls, "__dataclass_params__").frozen

            # Set non-init fields
            for f in fields(cls):
                if not f.init and f.name in data:
                    processed_value = process_attribute(data[f.name])
                    if is_frozen:
                        object.__setattr__(dataclass_instance, f.name, processed_value)
                    else:
                        setattr(dataclass_instance, f.name, processed_value)

            instance = cast(T, dataclass_instance)
        else:
            instance = cls.__new__(cls)
            for key, value in data.items():
                processed_value = process_attribute(value)
                setattr(instance, key, processed_value)

        return instance


def from_dict(dictionary: DictSerializableObject) -> DictSerializable:
    """Constructs an object instance conforming to the `DictSerializable` protocol from a dictionary.

    This function is a utility to deserialize a dictionary into an object that conforms to the
    `DictSerializable` protocol. It is typically used to convert data from JSON or similar formats
    back into complex Python objects.

    Args:
        dictionary (DictSerializableObject): A dictionary representing the serialized state of a DictSerializable object, containing a `type` key to identify the object class and an `attributes` key with its serialized attributes.

    Returns:
        DictSerializable: An instance of the class identified in the dictionary, populated with the provided attributes.
    """
    cls_name = dictionary["type"]
    cls_attributes = dictionary["attributes"]
    cls = DictSerializableFactory.get(cls_name)
    return cls.from_dict(cls_attributes)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains all the variables used inside a QProgram."""
from uuid import UUID, uuid4

from qililab.utils import DictSerializable, DictSerializableEnum


class Domain(DictSerializableEnum):
    """Domain class."""

    Scalar = (0,)
    Time = (1,)
    Frequency = (2,)
    Phase = (3,)
    Voltage = 4


class ValueSource(DictSerializableEnum):
    """ValueSource class"""

    Free = (0,)
    Dependent = 1


class Variable(DictSerializable):
    """Variable class used to define variables inside a QProgram."""

    @property
    def value(self):
        """Get the value of the variable

        Returns:
            int | float | None: The value of the variable
        """
        return self._value

    @property
    def domain(self):
        """Get the domain of the variable

        Returns:
            Domain: The domain of the variable
        """
        return self._domain

    def __init__(self, domain: Domain = Domain.Scalar) -> None:
        self._uuid: UUID = uuid4()
        self._source: ValueSource = ValueSource.Free
        self._value: int | float | None = None
        self._domain: Domain = domain

    def __repr__(self):
        return repr(self._uuid)

    def __hash__(self):
        return hash(self._uuid)

    def __eq__(self, other):
        return other is not None and isinstance(other, Variable) and self._uuid == other._uuid


class IntVariable(Variable, int):  # type: ignore
    """Integer variable. This class is used to define a variable of type int, such that Python recognizes this class
    as an integer."""

    def __new__(cls, _: Domain = Domain.Scalar):
        # Create a new float instance
        instance = int.__new__(cls, 0)
        return instance

    def __init__(self, domain: Domain = Domain.Scalar):
        Variable.__init__(self, domain)


class FloatVariable(Variable, float):  # type: ignore
    """Float variable. This class is used to define a variable of type float, such that Python recognizes this class
    as a float."""

    def __new__(cls, _: Domain = Domain.Scalar):
        # Create a new int instance
        instance = float.__new__(cls, 0.0)
        return instance

    def __init__(self, domain: Domain = Domain.Scalar):
        Variable.__init__(self, domain)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exponential decay correction."""
from copy import deepcopy
from dataclasses import dataclass

import numpy as np
from scipy import signal

from qililab.typings import PulseDistortionName
from qililab.utils import Factory

from .pulse_distortion import PulseDistortion


@Factory.register
@dataclass(frozen=True, eq=True)
class ExponentialCorrection(PulseDistortion):
    """Exponential decay distortion. Corrects an exponential decay using a linear IIR filter.

    Fitting should be done to y = g*(1+amp*exp(-t/tau)), where g is ignored in the corrections.

    For more info, check `SUPLEMENTAL MATERIAL IV. B. <https://arxiv.org/abs/1907.04818>`_.

    Args:
        tau_exponential (float): Tau exponential factor
        amp (float): Amplitude constant
        sampling_rate (float, optional): Sampling rate. Defaults to 1.
        norm_factor (float, optional): The manual normalization factor that multiplies the envelope in the apply() method. Defaults to 1 (no effect).
        auto_norm (bool, optional): Whether to automatically normalize the corrected envelope with the original max height in the apply() method.
            (The max height is the furthest number from 0 in the envelope, only checking the real axis/part). Defaults to True.

    Returns:
        PulseDistortion: Distortion to apply to given envelopes in PulseEvent.

    Examples:

        Imagine you want to distort a :class:`Rectangular` envelope with an `ExponentialCorrection`. You could do:

        >>> from qililab.pulse import Rectangular, BiasTeeCorrection
        >>> envelope = Rectangular().envelope(duration=50, amplitude=1.0)
        >>> distorted_envelope = ExponentialCorrection(tau_exponential=1.3, amp=2.0).apply(envelope)

        which would return a distorted envelope with the same real max height as the initial.

        >>> np.max(distorted_envelope) == np.max(envelope)
        True

        .. note::
            You can find more examples in the docstring of the :class:`PulseDistortion` base class.
    """

    name = (
        PulseDistortionName.EXPONENTIAL_CORRECTION
    )  #: Type of the correction. Enum type of PulseDistortionName class.
    tau_exponential: float  #: Tau exponential factor.
    amp: float  #: Amplitude constant. Value between 0 and 1.
    sampling_rate: float = 1.0  #: Sampling rate. Defaults to 1.

    def apply(self, envelope: np.ndarray) -> np.ndarray:
        """Distorts envelopes (originally created to distort square envelopes).

        Corrects an exponential decay using a linear IIR filter.

        Fitting should be done to y = g*(1+amp*exp(-t/tau)), where g is ignored in the corrections.

        If `self.auto_norm` is True (default) normalizes the resulting envelope to have the same real max height than the starting one.
        (the max height is the furthest number from 0, only checking the real axis/part)
        If the corrected envelope is zero everywhere or doesn't have a real part this process is skipped.

        Finally it applies the manual `self.norm_factor` to the result, reducing the full envelope by its magnitude.

        For further details on the normalization implementation see the docstring on :class:`PulseDistortion` base class.

        Args:
            envelope (numpy.ndarray): Array representing the envelope of a pulse for each time step.

        Returns:
            numpy.ndarray: Amplitude of the envelope for each time step.
        """
        if self.amp >= 0.0:
            # Parameters
            alpha = 1 - np.exp(-1 / (self.sampling_rate * self.tau_exponential * (1 + self.amp)))
            k = self.amp / (1 + self.amp - alpha)

            # Coefficients
            a_1 = -(1 - alpha)
            b_0 = 1 - k + k * alpha
            b_1 = -(1 - k) * (1 - alpha)

        else:
            # Parameters
            denominator = 2 * self.tau_exponential * (1 + self.amp) + 1

            # Coefficients
            a_1 = (self.tau_exponential * (1 + 2 * self.amp) - 1) / denominator
            b_0 = (2 * self.tau_exponential + 1) / denominator
            b_1 = (-2 * self.tau_exponential + 1) / denominator

        # Filtered signal
        corr_envelope = signal.lfilter(b=[b_0, b_1], a=[1, a_1], x=envelope)
        return self.normalize_envelope(envelope=envelope, corr_envelope=corr_envelope)

    @classmethod
    def from_dict(cls, dictionary: dict) -> "ExponentialCorrection":
        """Loads ExponentialCorrection object from dictionary.

        Args:
            dictionary (dict): Dictionary object of the ExponentialCorrection object. It must include the name of the
            correction, the tau exponential factor, the amplitude, the sampling rate, the normalization factor and
            the auto normalization flag value.

        Returns:
            ExponentialCorrection: Loaded class.
        """
        local_dictionary = deepcopy(dictionary)
        local_dictionary.pop("name", None)
        return cls(**local_dictionary)

    def to_dict(self) -> dict:
        """Returns dictionary representation of the distortion.

        Returns:
            dict: Dictionary representation including the name of the correction, the tau exponential factor, the
            amplitude, the sampling rate, the normalization factor and the auto normalization flag value.
        """
        return {
            "name": self.name.value,
            "tau_exponential": self.tau_exponential,
            "amp": self.amp,
            "sampling_rate": self.sampling_rate,
            "norm_factor": self.norm_factor,
            "auto_norm": self.auto_norm,
        }



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" AWG Qblox ADC Sequencer """
from dataclasses import dataclass

from qililab.instruments.awg_settings.awg_adc_sequencer import AWGADCSequencer
from qililab.instruments.awg_settings.awg_qblox_sequencer import AWGQbloxSequencer


@dataclass
class AWGQbloxADCSequencer(AWGQbloxSequencer, AWGADCSequencer):
    """AWG Qblox ADC Sequencer"""

    qubit: int
    weights_i: list[float]
    weights_q: list[float]
    weighed_acq_enabled: bool

    def __post_init__(self):
        self._verify_weights()

    def _verify_weights(self):
        """Verifies that the length of weights_i and weights_q are equal.

        Raises:
            IndexError: The length of weights_i and weights_q must be equal.
        """
        if len(self.weights_i) != len(self.weights_q):
            raise IndexError("The length of weights_i and weights_q must be equal.")

    @property
    def used_integration_length(self) -> int:
        """Final integration length used by the AWG in the integration.

        Returns:
            int: Length of the weights if weighed acquisition is enabled, configured `integration_length` if disabled.
        """
        return len(self.weights_i) if self.weighed_acq_enabled else self.integration_length



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Comparison models, to use in ``CalibrationController`` ``check_data()``."""
from typing import Any

import numpy as np

# Epsilon, for avoiding divisions by zero.
div_epsilon = 0.00000000000000000000000000001


def norm_root_mean_sqrt_error(obtained: dict[str, list], comparison: dict[str, list]) -> float:
    """Returns the normalized RMSE (mean absolute error) between the comparison and obtained samples.

    If the comparison has a fit, it will be used, if not, the results will.

    The samples have to have the following structure (lists can be arrays as well):

    obtained/comparison = {
        "sweep_interval": list_of_the_sweep,
        "results": list_of_results
        "fit": list_of_fitting,  # optional
    }

    Args:
        obtained (dict): obtained samples to compare. Structure following the function docstring.
        comparison (dict): previous samples to compare. Structure following the function docstring.

    Returns:
        float: difference/error between the two samples.
    """
    (fit, check) = (True, "fit") if "fit" in comparison else (False, "results")

    # Structure checks:
    is_structure_of_check_parameters_correct(obtained, comparison, fit)

    for check_data in [obtained, comparison]:
        if not np.shape(check_data[check]) == np.shape(check_data["results"]) == (len(check_data["sweep_interval"]),):
            raise ValueError("Incorrect 'check_parameters' shape for this notebook output.")

        if 0 in [len(check_data["sweep_interval"]), len(check_data["results"])]:
            raise ValueError("Empty 'sweep_interval' or 'results' in `check_parameters`'s notebook output.")

    # Error computation:
    # TODO: Avoid comparison "if obtained_x in comparison["sweep_interval"]" instead build comparison sweep interval taking into acount this
    # TODO: If no overlapping sweep intervals we must return a huge error so we get bad_data directly
    square_error = sum(
        (obtained["results"][i] - comparison[check][comparison["sweep_interval"].index(obtained_x)]) ** 2
        for i, obtained_x in enumerate(obtained["sweep_interval"])
        if obtained_x in comparison["sweep_interval"]
    )
    root_mean_square_error = np.sqrt(square_error / len(obtained["results"]))

    return root_mean_square_error / np.abs(np.mean(comparison[check]) + div_epsilon)


def IQ_norm_root_mean_sqrt_error(obtained: dict[str, list], comparison: dict[str, list]) -> float:
    """Returns the normalized RMSE (mean absolute error) between the comparison and obtained samples for I and Q.

    If the comparison has a fit, it will be used, if not, the results will.

    The samples have to have the following structure (lists can be arrays as well):

    obtained/comparison = {
        "sweep_interval": list_of_the_sweep,
        "results": [list_of_I_results, list_of_Q_results],
        "fit": [list_of_I_fitting, list_of_Q_fitting],  # optional
    }

    Args:
        obtained (dict): obtained samples to compare. Structure following the function docstring.
        comparison (dict): previous samples to compare. Structure following the function docstring.

    Returns:
        float: difference/error between the two samples.
    """
    (fit, check) = (True, "fit") if "fit" in comparison else (False, "results")

    # Structure checks:
    is_structure_of_check_parameters_correct(obtained, comparison, fit)

    for check_data in [obtained, comparison]:
        if not np.shape(check_data[check]) == np.shape(check_data["results"]) == (2, len(check_data["sweep_interval"])):
            raise ValueError("Incorrect 'check_parameters' shape for this notebook output.")

        if 0 in [len(check_data["sweep_interval"]), len(check_data["results"][0]), len(check_data["results"][1])]:
            raise ValueError("Empty 'sweep_interval' or 'results' in `check_parameters`'s notebook output.")

    # Error computation:
    i, q = obtained["results"]
    errors = []
    # TODO: Avoid comparison "if obtained_x in comparison["sweep_interval"]" instead build comparison sweep interval taking into acount this
    # TODO: If no overlapping sweep intervals we must return a huge error so we get bad_data directly
    for idx, obtained_results in enumerate([i, q]):
        square_error = sum(
            (obtained_results[index] - comparison[check][idx][comparison["sweep_interval"].index(obtained_x)]) ** 2
            for index, obtained_x in enumerate(obtained["sweep_interval"])
            if obtained_x in comparison["sweep_interval"]
        )
        root_mean_square_error = np.sqrt(square_error / len(obtained_results))

        # normalize the difference with the mean values, and add it to the i or q error
        errors.append(root_mean_square_error / np.abs(np.mean(comparison[check]) + div_epsilon))

    # Return the one with less errors, since it was the one used for the fitting.
    return np.min(errors)


# pylint: disable=too-many-locals
def ssro_comparison_2D(obtained: dict[str, Any], comparison: dict[str, Any]) -> float:
    """Returns a normalized error between the comparison and obtained samples, for 2D plots.

    Always compares results vs results, no fit used.

    The samples have to have the following structure (lists can be arrays as well):

    obtained/comparison = {
        "sweep_interval": NUM_BINS,
        "results": ((list_i_0, list_i_1), (list_q_0, list_q_1)),  # the index indicates the gaussian.
    }

    Args:
        obtained (dict): Obtained samples to compare. Structure following the function docstring.
        comparison (dict): pPrevious samples to compare. Structure following the function docstring.

    Returns:
        float: difference/error between the two samples.
    """
    # Structure checks:
    for check_data in [obtained, comparison]:
        if "sweep_interval" not in check_data or "results" not in check_data:
            raise ValueError("Keys in the `check_parameters` are not 'sweep_interval' and 'results', as is needed.")

        if np.shape(check_data["results"]) != (2, 2, check_data["sweep_interval"]):
            raise ValueError("Incorrect 'check_parameters' shape for this notebook output.")

        for i in range(2):
            if len(check_data["results"][0][i]) == 0 or len(check_data["results"][1][i]) == 0:
                raise ValueError("Empty 'sweep_interval' or 'results' in `check_parameters`'s notebook output.")

    i_s: list[list[list[list]]] = [[[], []], [[], []]]
    q_s: list[list[list[list]]] = [[[], []], [[], []]]
    mean_diff, std_dev_diff = 0, 0
    # Error computation:
    for gaussian_n in range(2):  # For first and secong gaussian
        for datas, data in enumerate([obtained, comparison]):
            i_s[gaussian_n][datas] = data["results"][0][datas]  # 0 for i's, and j for first or second gaussian
            q_s[gaussian_n][datas] = data["results"][1][datas]  # 1 for q's, and j for first or second gaussian

        # Difference in means of 2d histograms
        mean_diff += (
            (np.mean(i_s[gaussian_n][0]) - np.mean(i_s[gaussian_n][1])) / (np.mean(i_s[gaussian_n][0]) + div_epsilon)
        ) ** 2 + (
            (np.mean(q_s[gaussian_n][0]) - np.mean(q_s[gaussian_n][1])) / (np.mean(q_s[gaussian_n][0]) + div_epsilon)
        ) ** 2

        # Difference in std of 2d histograms
        std_dev_diff += (
            (np.std(i_s[gaussian_n][0]) - np.std(i_s[gaussian_n][1])) / (np.std(i_s[gaussian_n][0]) + div_epsilon)
        ) ** 2 + (
            (np.std(q_s[gaussian_n][0]) - np.std(q_s[gaussian_n][1])) / (np.std(q_s[gaussian_n][0]) + div_epsilon)
        ) ** 2

    # Return a weighted sum, with more weight in the average than the standard deviation:
    return np.sqrt((mean_diff) * 4 + (std_dev_diff) / (obtained["sweep_interval"] + div_epsilon)) / 10


def is_structure_of_check_parameters_correct(obtained: dict[str, list], comparison: dict[str, list], fit: bool = True):
    """If the structure is incorrect it will raise an error.

    Args:
        obtained (dict): obtained samples to compare.
        comparison (dict): previous samples to compare.
        fit (bool): flag, to wether compare against the previous fit as well as the previous results. Defaults to True (against fit).

    Raises:
        ValueError: if keys don't match the expected.
    """
    for check_data in [obtained, comparison]:
        if "sweep_interval" not in check_data or "results" not in check_data or (fit and "fit" not in check_data):
            raise ValueError(
                "Keys in the `check_parameters` are not 'sweep_interval', 'results' (and 'fit', optional), as is needed."
            )



"""
This module contains all the methods and classes used for automatic calibration.

A high level view of how the automatic calibration algorithm works and how the different classes and methods are used
can be read from: https://arxiv.org/abs/1803.03226

**Examples** are found in the `qililab/examples/calibration` folder

.. currentmodule:: qililab.calibration

CalibrationController class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.. autosummary::
    :toctree: api

    ~CalibrationController

CalibrationNode Class
~~~~~~~~~~~~~~~~~~~~~~
.. autosummary::
    :toctree: api

    ~CalibrationNode

Calibration-related methods
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.. autosummary::
    :toctree: api

    ~export_nb_outputs
"""

from .calibration_controller import CalibrationController
from .calibration_node import CalibrationNode, export_nb_outputs



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantum Machines Manager class."""
import os
from dataclasses import dataclass
from typing import Any, Dict, cast

import numpy as np
from qm import DictQuaConfig, QuantumMachine, QuantumMachinesManager, SimulationConfig
from qm.jobs.running_qm_job import RunningQmJob
from qm.octave import QmOctaveConfig
from qm.qua import Program

from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import InstrumentName, Parameter, QMMDriver
from qililab.utils import hash_qua_program, merge_dictionaries


@InstrumentFactory.register
class QuantumMachinesCluster(Instrument):
    """Class defining the Qililab Quantum Machines Cluster instrument in Qililab.

    This class allows Qililab control and communication with an instance of the
    Quantum Machines Manager, which is the central class to interact with Quantum Machines instruments.
    The manager is responsible, through the use of QUA sequences, of setting the Quantum Machines instruments
    in the right manner for Quantum Control.

    Args:
        name (InstrumentName): Name of the Instrument.
        device (QMMDriver): Instance of the Quantum Machines Manager Driver class.
        settings (QMMSettings): Settings of the instrument.
    """

    name = InstrumentName.QUANTUM_MACHINES_CLUSTER

    @dataclass
    class QuantumMachinesClusterSettings(Instrument.InstrumentSettings):
        """Settings for Quantum Machines Cluster instrument.

        Args:
            address (str): IP address to connect to the Quantum Machines instruments.
            cluster (str): Name of the cluster to connect to.
            run_octave_calibration (bool): Flag to indicate if octave calibration should be run on connecting to Quantum Machines.
            octaves (list[dict[str, Any]]): List of octaves the quantum machines stack has.
            controllers (list[dict[str, Any]]): List of controllers (instruments) the quantum machines stack has.
            elements (list[dict[str, Any]]): List of elements (buses) the quantum machines stack has.
        """

        address: str
        cluster: str
        run_octave_calibration: bool
        octaves: list[dict[str, Any]]
        controllers: list[dict[str, Any]]
        elements: list[dict[str, Any]]

        def to_qua_config(self) -> Dict[str, Any]:
            """Creates the Quantum Machines QUA config dictionary.

            Creates, an instance of a dictionary in the format that QuantumMachines expects the config dictionary to be.

            The values of following keys of the dictionary are parsed from QMMSettings:
                - controllers (OPX+ instruments)
                - octaves (Octave instruments)
                - mixers (up/down conversion settings if not using octaves)
                - elements (buses)

            The values of the rest keys are empty, meant to be updated in runtime.

            Returns:
                config: Dict[str, Any]
            """
            elements, mixers = self._get_elements_and_mixers_config()
            controllers = self._get_controllers_config()
            octaves = self._get_octaves_config()

            return {
                "version": 1,
                "controllers": controllers,
                "elements": elements,
                "mixers": mixers,
                "waveforms": {},
                "integration_weights": {},
                "pulses": {},
                "digital_waveforms": {},
                "octaves": octaves,
            }

        def _get_controllers_config(self) -> dict[str, Any]:
            """Returns the controllers config dictionary.

            Returns:
                controllers: Dict[str, Any]
            """
            return {
                controller["name"]: {
                    "analog_outputs": {
                        output["port"]: {
                            "offset": output["offset"] if "offset" in output else 0.0,
                            "delay": output["delay"] if "delay" in output else 0.0,
                        }
                        for output in controller.get("analog_outputs", [])
                    },
                    "analog_inputs": {
                        input["port"]: {
                            "offset": input["offset"] if "offset" in input else 0.0,
                            "gain_db": input["gain"] if "gain" in input else 0.0,
                        }
                        for input in controller.get("analog_inputs", [])
                    },
                    "digital_outputs": {output["port"]: {} for output in controller.get("digital_outputs", [])},
                }
                for controller in self.controllers
            }

        def _get_octaves_config(self) -> dict[str, Any]:
            """Returns the octaves config dictionary.

            Returns:
                octaves: Dict[str, Any]
            """
            return {
                octave["name"]: {
                    "RF_outputs": {
                        output["port"]: {
                            "LO_frequency": output["lo_frequency"],  # Should be between 2 and 18 GHz.
                            "LO_source": "internal",
                            "gain": output["gain"] if "gain" in output else 0.0,
                            "output_mode": "always_on",
                            "input_attenuators": "OFF",  # can be: "OFF" / "ON". Default is "OFF".
                        }
                        for output in octave.get("rf_outputs", [])
                    },
                    "RF_inputs": {
                        output["port"]: {
                            "RF_source": "RF_in",
                            "LO_frequency": output["lo_frequency"],
                            "LO_source": "internal",  # can be: "internal" / "external". Default is "internal".
                            "IF_mode_I": "direct",  # can be: "direct" / "mixer" / "envelope" / "off". Default is "direct".
                            "IF_mode_Q": "direct",
                        }
                        for output in octave.get("rf_inputs", [])
                    },
                    "connectivity": octave["controller"],
                }
                for octave in self.octaves
            }

        def _get_elements_and_mixers_config(self) -> tuple:
            """Returns the elements config dictionary.

            Returns:
                elements: Dict[str, Any]
            """
            elements = {}
            mixers = {}

            for element in self.elements:
                bus_name = element["bus"]
                element_dict: dict = {"operations": {}}

                # Flux bus
                if "single_input" in element:
                    element_dict["singleInput"] = {
                        "port": (element["single_input"]["controller"], element["single_input"]["port"]),
                    }
                # IQ bus
                elif "mix_inputs" in element:
                    mixer_name = f"mixer_{bus_name}"
                    lo_frequency = int(element["mix_inputs"]["lo_frequency"])
                    intermediate_frequency = int(element["intermediate_frequency"])
                    mixer_correction = (
                        element["mix_inputs"]["mixer_correction"]
                        if "mixer_correction" in element["mix_inputs"]
                        else [1.0, 0.0, 0.0, 1.0]
                    )
                    mixers[mixer_name] = [
                        {
                            "intermediate_frequency": intermediate_frequency,
                            "lo_frequency": lo_frequency,
                            "correction": mixer_correction,
                        }
                    ]
                    element_dict["mixInputs"] = {
                        key: (element["mix_inputs"][key]["controller"], element["mix_inputs"][key]["port"])
                        for key in ["I", "Q"]
                    } | {"lo_frequency": lo_frequency, "mixer": mixer_name}
                    element_dict["intermediate_frequency"] = intermediate_frequency

                    # readout bus
                    if "outputs" in element:
                        element_dict["outputs"] = {
                            key: (element["outputs"][key]["controller"], element["outputs"][key]["port"])
                            for key in ["out1", "out2"]
                        }
                        element_dict["time_of_flight"] = element["time_of_flight"]
                        element_dict["smearing"] = element["smearing"]
                # RF with Octave
                elif "rf_inputs" in element:
                    element_dict["RF_inputs"] = {"port": (element["rf_inputs"]["octave"], element["rf_inputs"]["port"])}
                    element_dict["intermediate_frequency"] = int(element["intermediate_frequency"])

                    # readout bus
                    if "rf_outputs" in element:
                        element_dict["RF_outputs"] = {
                            "port": (element["rf_outputs"]["octave"], element["rf_outputs"]["port"]),
                        }
                        element_dict["time_of_flight"] = element["time_of_flight"]
                        element_dict["smearing"] = element["smearing"]

                # other settings
                if "digital_inputs" in element:
                    element_dict["digitalInputs"] = {
                        "switch": {
                            "port": (element["digital_inputs"]["controller"], element["digital_inputs"]["port"]),
                            "delay": element["digital_inputs"]["delay"],
                            "buffer": element["digital_inputs"]["buffer"],
                        }
                    }
                if "digital_outputs" in element:
                    element_dict["digitalOutputs"] = {
                        "port": (element["digital_outputs"]["controller"], element["digital_outputs"]["port"]),
                    }

                elements[bus_name] = element_dict

            return elements, mixers

    settings: QuantumMachinesClusterSettings
    device: QMMDriver
    _qmm: QuantumMachinesManager
    _qm: QuantumMachine
    _config: DictQuaConfig
    _octave_config: QmOctaveConfig | None = None
    _is_connected_to_qm: bool = False
    _compiled_program_cache: dict[str, str] = {}

    @property
    def config(self) -> DictQuaConfig:
        """Get the QUA config dictionary."""
        return self._config

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """Sets initial instrument settings.

        Creates an instance of the Qilililab Quantum Machines Manager, and sets the configuration dictionary.
        """
        if self.settings.octaves:
            self._octave_config = QmOctaveConfig()
            self._octave_config.set_calibration_db(os.getcwd())
            for octave in self.settings.octaves:
                self._octave_config.add_device_info(octave["name"], self.settings.address, octave["port"])
        self._qmm = QuantumMachinesManager(
            host=self.settings.address, cluster_name=self.settings.cluster, octave=self._octave_config
        )

        self._config = self.settings.to_qua_config()

    @Instrument.CheckDeviceInitialized
    def turn_on(self):
        """Turns on the instrument."""
        if not self._is_connected_to_qm:
            self._qm = self._qmm.open_qm(config=self._config, close_other_machines=True)
            self._compiled_program_cache = {}
            self._is_connected_to_qm = True

            if self.settings.run_octave_calibration:
                self.run_octave_calibration()

    @Instrument.CheckDeviceInitialized
    def reset(self):
        """Resets instrument settings."""

    @Instrument.CheckDeviceInitialized
    def turn_off(self):
        """Turns off an instrument."""
        if self._is_connected_to_qm:
            self._qm.close()
            self._is_connected_to_qm = False

    def append_configuration(self, configuration: dict):
        """Update the configuration dictionary by appending the configuration generated by compilation."""
        merged_configuration = merge_dictionaries(dict(self._config), configuration)
        if self._config != merged_configuration:
            self._config = cast(DictQuaConfig, merged_configuration)
            # If we are already connected, reopen the connection with the new configuration
            if self._is_connected_to_qm:
                self._qm = self._qmm.open_qm(config=self._config, close_other_machines=True)
                self._compiled_program_cache = {}

    def run_octave_calibration(self):
        """Run calibration procedure for the buses with octaves, if any."""
        elements = [element for element in self._config["elements"] if "RF_inputs" in self._config["elements"][element]]
        for element in elements:
            self._qm.calibrate_element(element)

    def set_parameter_of_bus(self, bus: str, parameter: Parameter, value: float | str | bool):
        """Sets the parameter of the instrument

        Args:
            bus (str): The assossiated bus to change parameter
            parameter (Parameter): The parameter to update
            value (float | str | bool): The new value of the parameter

        Raises:
            NotImplementedError: Raised if not connected to Quantum Machines
            ParameterNotFound: Raised if parameter does not exist
        """
        # TODO: Change private QM API to public when implemented.
        if not self._is_connected_to_qm:
            raise NotImplementedError(f"You should be connected to {self.name} in order to change a parameter.")

        element = next((element for element in self.settings.elements if element["bus"] == bus), None)
        if element is None:
            raise ValueError(f"Bus {bus} was not found in {self.name} settings.")

        if parameter in [Parameter.LO_FREQUENCY, Parameter.GAIN]:
            if "rf_inputs" not in element:
                raise ValueError(
                    f"Trying to change parameter {parameter.name} in {self.name}, however bus {bus} is not connected to an octave."
                )
            octave_name = element["rf_inputs"]["octave"]
            out_port = element["rf_inputs"]["port"]
            in_port = element["rf_outputs"]["port"] if "rf_outputs" in element else None
            settings_octave = next(octave for octave in self.settings.octaves if octave["name"] == octave_name)
            settings_octave_rf_output = next(
                rf_output for rf_output in settings_octave["rf_outputs"] if rf_output["port"] == out_port
            )

            if parameter == Parameter.LO_FREQUENCY:
                lo_frequency = float(value)
                self._qm.octave.set_lo_frequency(element=bus, lo_frequency=lo_frequency)
                settings_octave_rf_output["lo_frequency"] = lo_frequency
                self._config["octaves"][octave_name]["RF_outputs"][out_port]["LO_frequency"] = lo_frequency

                if in_port is not None:
                    settings_octave_rf_input = next(
                        rf_input for rf_input in settings_octave["rf_inputs"] if rf_input["port"] == in_port
                    )
                    settings_octave_rf_input["lo_frequency"] = lo_frequency
                    self._config["octaves"][octave_name]["RF_inputs"][in_port]["LO_frequency"] = lo_frequency
                return
            if parameter == Parameter.GAIN:
                gain_in_db = float(value)
                self._qm.octave.set_rf_output_gain(element=bus, gain_in_db=gain_in_db)
                settings_octave_rf_output["gain"] = gain_in_db
                self._config["octaves"][octave_name]["RF_outputs"][out_port]["gain"] = gain_in_db
                return
        if parameter == Parameter.IF:
            intermediate_frequency = float(value)
            self._qm.set_intermediate_frequency(element=bus, freq=intermediate_frequency)
            element["intermediate_frequency"] = intermediate_frequency
            self._config["elements"][bus]["intermediate_frequency"] = intermediate_frequency
            return
        raise ParameterNotFound(f"Could not find parameter {parameter} in instrument {self.name}.")

    def get_parameter_of_bus(self, bus: str, parameter: Parameter):
        """Gets the value of a parameter

        Args:
            bus (str): The assossiated bus of the parameter
            parameter (Parameter): The parameter to get value

        Raises:
            ParameterNotFound: Raised if parameter does not exist
        """
        # TODO: Change private QM API to public when implemented.
        if parameter == Parameter.LO_FREQUENCY:
            return self._qm._elements[bus].input.lo_frequency  # type: ignore[union-attr] # pylint: disable=protected-access
        if parameter == Parameter.GAIN:
            return self._qm._elements[bus].input.gain  # type: ignore[union-attr] # pylint: disable=protected-access
        if parameter == Parameter.IF:
            return self._qm._elements[bus].intermediate_frequency  # pylint: disable=protected-access
        raise ParameterNotFound(f"Could not find parameter {parameter} in instrument {self.name}")

    def compile(self, program: Program) -> str:
        """Compiles and stores a given QUA program on the Quantum Machines instance,
        and returns a unique identifier associated with the compiled program.

        The method first generates a hash for the input QUA program. If this hash is not already present in
        the cache, it proceeds to compile the program using QM's compile method and stores the result in the cache
        indexed by the generated hash. This caching mechanism prevents recompiling the same program multiple times,
        thus optimizing performance.

        Args:
            program (Program): The QUA program to be compiled.

        Returns:
            str: A unique identifier (hash) for the compiled QUA program. This identifier can be used to retrieve the compiled program from the cache, or run it  with `run_compiled_program` method.
        """
        qua_program_hash = hash_qua_program(program=program)
        if qua_program_hash not in self._compiled_program_cache:
            self._compiled_program_cache[qua_program_hash] = self._qm.compile(program=program)
        return self._compiled_program_cache[qua_program_hash]

    def run_compiled_program(self, compiled_program_id: str) -> RunningQmJob:
        """Executes a previously compiled QUA program identified by its unique compiled program ID.

        This method submits the compiled program to the Quantum Machines (QM) execution queue and waits for
        its execution to complete. The execution of the program is managed by the QM's queuing system,
        which schedules and runs the job on the quantum hardware or simulator as per availability and queue status.

        Args:
            compiled_program_id (str): The unique identifier of the compiled QUA program to be executed. This ID should correspond to a program that has already been compiled and is present in the program cache.

        Returns:
            RunningQmJob: An object representing the running job. This object provides methods and properties to check the status of the job, retrieve results upon completion, and manage or investigate the job's execution.
        """
        pending_job = self._qm.queue.add_compiled(compiled_program_id)
        return pending_job.wait_for_execution()

    def run(self, program: Program) -> RunningQmJob:
        """Runs the QUA Program.

        Creates, for every execution, an instance of a QuantumMachine class and executes the QUA program on it.
        After execution, the job generated by the Quantum Machines Manager is returned.

        Args:
            program (Program): QUA Program to be run on Quantum Machines instruments.

        Returns:
            job: Quantum Machines job.
        """

        return self._qm.execute(program)

    def get_acquisitions(self, job: RunningQmJob) -> dict[str, np.ndarray]:
        """Fetches the results from the execution of a QUA Program.

        Once the results have been fetched, they are returned wrapped in a QuantumMachinesResult instance.

        Args:
            job (QmJob): Job that provides the result handles.

        Returns:
            QuantumMachinesResult: Quantum Machines result instance.
        """
        result_handles_fetchers = job.result_handles
        result_handles_fetchers.wait_for_all_values()
        results = {
            name: handle.fetch_all(flat_struct=True) for name, handle in job.result_handles if handle is not None
        }
        return {name: data for name, data in results.items() if data is not None}

    def simulate(self, program: Program) -> RunningQmJob:
        """Simulates the QUA Program.

        Creates, for every simulation, an instance of a QuantumMachine class and executes the QUA program on it.
        It returns the running job instance.

        Args:
            program (Program): QUA Program to be simulated on Quantum Machines instruments.

        Returns:
            job: Quantum Machines job.
        """
        return self._qm.simulate(program, SimulationConfig(40_000))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QDevil QDAC-II Instrument"""

from dataclasses import dataclass

from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.instruments.voltage_source import VoltageSource
from qililab.typings import InstrumentName
from qililab.typings import QDevilQDac2 as QDevilQDac2Driver
from qililab.typings.enums import Parameter


@InstrumentFactory.register
class QDevilQDac2(VoltageSource):
    """QDevil QDAC-II Intrument

    Args:
        name (InstrumentName): Name of the instrument
        device (QDevilQDac2): Instance of the QCoDes QDac2 class.
        settings (QDevilQDac2Settings): Settings of the instrument.
    """

    name = InstrumentName.QDEVIL_QDAC2

    @dataclass
    class QDevilQDac2Settings(VoltageSource.VoltageSourceSettings):
        """Contains the settings of a specific signal generator."""

        low_pass_filter: list[str]

    settings: QDevilQDac2Settings
    device: QDevilQDac2Driver

    @property
    def low_pass_filter(self):
        """QDAC-II `low_pass_filter` property.

        Returns:
            list[str]: A list of the low pass filter setting of each DACs.
        """
        return self.settings.low_pass_filter

    def setup(  # pylint: disable=too-many-branches
        self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None
    ):
        """Set parameter to the corresponding value for an instrument's channel.

        Args:
            parameter (Parameter): Name of the parameter to be updated
            value (float | str | bool): New value of the parameter
            channel_id (int | None): Channel identifier
        """
        self._validate_channel(channel_id=channel_id)

        channel = self.device.channel(channel_id) if self.is_device_active() else None

        index = self.dacs.index(channel_id)
        if parameter == Parameter.VOLTAGE:
            voltage = float(value)
            self.settings.voltage[index] = voltage
            if self.is_device_active():
                channel.dc_constant_V(voltage)
            return
        if parameter == Parameter.SPAN:
            span = str(value)
            self.settings.span[index] = span
            if self.is_device_active():
                channel.output_range(span)
            return
        if parameter == Parameter.RAMPING_ENABLED:
            ramping_enabled = bool(value)
            self.settings.ramping_enabled[index] = ramping_enabled
            if self.is_device_active():
                if ramping_enabled:
                    channel.dc_slew_rate_V_per_s(self.ramp_rate[index])
                else:
                    channel.dc_slew_rate_V_per_s(2e7)
            return
        if parameter == Parameter.RAMPING_RATE:
            ramping_rate = float(value)
            self.settings.ramp_rate[index] = ramping_rate
            ramping_enabled = self.ramping_enabled[index]
            if ramping_enabled and self.is_device_active():
                channel.dc_slew_rate_V_per_s(ramping_rate)
            return
        if parameter == Parameter.LOW_PASS_FILTER:
            low_pass_filter = str(value)
            self.settings.low_pass_filter[index] = low_pass_filter
            if self.is_device_active():
                channel.output_filter(low_pass_filter)
            return
        raise ParameterNotFound(f"Invalid Parameter: {parameter.value}")

    def get(self, parameter: Parameter, channel_id: int | None = None):
        """Get parameter's value for an instrument's channel.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier.
        """
        self._validate_channel(channel_id=channel_id)

        index = self.dacs.index(channel_id)
        if hasattr(self.settings, parameter.value):
            return getattr(self.settings, parameter.value)[index]
        raise ParameterNotFound(f"Could not find parameter {parameter} in instrument {self.name}")

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """Perform an initial setup."""
        for channel_id in self.dacs:
            self._validate_channel(channel_id=channel_id)

            index = self.dacs.index(channel_id)
            channel = self.device.channel(channel_id)
            channel.dc_mode("fixed")
            channel.output_range(self.span[index])
            channel.output_filter(self.low_pass_filter[index])
            if self.ramping_enabled[index]:
                channel.dc_slew_rate_V_per_s(self.ramp_rate[index])
            else:
                channel.dc_slew_rate_V_per_s(2e7)
            channel.dc_constant_V(0.0)

    @Instrument.CheckDeviceInitialized
    def turn_on(self):
        """Start outputing voltage."""
        for channel_id in self.dacs:
            index = self.dacs.index(channel_id)
            channel = self.device.channel(channel_id)
            channel.dc_constant_V(self.voltage[index])

    @Instrument.CheckDeviceInitialized
    def turn_off(self):
        """Stop outputing voltage."""
        for channel_id in self.dacs:
            channel = self.device.channel(channel_id)
            channel.dc_constant_V(0.0)

    @Instrument.CheckDeviceInitialized
    def reset(self):
        """Reset instrument. This will affect all channels."""
        self.device.reset()

    def _validate_channel(self, channel_id: int | None):
        """Check if channel identifier is valid and in the allowed range."""
        if channel_id is None:
            raise ValueError(
                "QDevil QDAC-II is a multi-channel instrument. `channel_id` must be specified to get or set parameters."
            )
        if channel_id < 1 or channel_id > 24:
            raise ValueError(f"The specified `channel_id`: {channel_id} is out of range. Allowed range is [1, 24].")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox Pulsar Controller class"""
from dataclasses import dataclass

from qililab.instrument_controllers.qblox.qblox_controller import QbloxController
from qililab.instrument_controllers.single_instrument_controller import SingleInstrumentController
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.typings.enums import ConnectionName, InstrumentControllerName
from qililab.typings.instruments.pulsar import Pulsar


@InstrumentControllerFactory.register
class QbloxPulsarController(SingleInstrumentController, QbloxController):
    """Qblox Pulsar Controller class.

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        settings (QbloxPulsarControllerSettings): Settings of the Qblox Pulser Instrument Controller.
    """

    name = InstrumentControllerName.QBLOX_PULSAR
    device: Pulsar

    @dataclass
    class QbloxPulsarControllerSettings(QbloxController.QbloxControllerSettings):
        """Contains the settings of a specific Qblox Pulsar Controller."""

        def __post_init__(self):
            super().__post_init__()
            self.connection.name = ConnectionName.TCP_IP

    settings: QbloxPulsarControllerSettings

    def _initialize_device(self):
        """Initialize device controller."""
        self.device = Pulsar(name=f"{self.name.value}_{self.alias}", identifier=self.address)

    @QbloxController.CheckConnected
    def _set_reference_source(self):
        """Set reference source. Options are 'internal' or 'external'"""
        self.modules[0].device.reference_source(self.reference_clock.value)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Loop class."""
from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np

from qililab.constants import LOOP
from qililab.typings.enums import Parameter


@dataclass
class Loop:
    """Class used to loop a parameter over the given array values.

    Args:
        alias (str): Alias of the object holding the parameter to loop over.
        parameter (Parameter): Parameter to loop.
        values (ndarray): Array of values to loop over.
        loop (Loop | None): Inner loop. If not None, a nested loop is created. Defaults to None.

    **Examples**:
    Here is an example to create a loop over the frequency of the instrument 'AWG' with the given values:
    >>> loop = Loop(alias='AWG', parameter=Parameter.FREQUENCY, values=np.linspace(7e9, 8e9, num=10))

    Any array of values can be used. For example, one can use `np.logsapce` or `np.geomspace` to create a logarithmic
    loop:
    >>> loop = Loop(alias='AWG', parameter=Parameter.FREQUENCY, values=np.logspace(7e9, 8e9, num=10))

    The difference between `np.logspace` or `np.geomspace` is that geomspace specifies the exact start and stop
    points while logspace specifies the exponent of the start and stop given a base. See below:
    np.logspace(start=1, stop=5, num=5, base=2) -> [base**start .. base**stop] -> [2,4,8,16,32]
    np.geomspace(start=2, stop=32, num=5) -> [start .. stop] -> [2,4,8,16,32]

    One can also create nested loops:
    >>> inner_loop = Loop(alias='AWG', parameter=Parameter.FREQUENCY, values=np.arange(7e9, 8e9, step=10))
    >>> outer_loop = Loop(alias='AWG', parameter=Parameter.POWER, values=np.linspace(0, 10, num=10), loop=inner_loop)
    """

    alias: str
    parameter: Parameter
    values: np.ndarray
    loop: Loop | None = None
    previous: Loop | None = field(compare=False, default=None)
    channel_id: int | None = None

    def __post_init__(self):
        """Check that either step or num is used. Overwrite 'previous' attribute of next loop with self."""
        if self.loop is not None:
            if isinstance(self.loop, dict):
                self.loop = Loop(**self.loop)
            self.loop.previous = self
        if isinstance(self.parameter, str):
            self.parameter = Parameter(self.parameter)
        if isinstance(self.values, list):
            self.values = np.array(self.values, dtype=object)

    @property
    def all_values(self) -> np.ndarray:
        """Loop 'all_values' property.

        Returns:
            list: Values of all loops.
        """
        all_values = [loop.values for loop in self.loops]
        return np.array(all_values, dtype=object)

    @property
    def shape(self) -> list[int]:
        """Return number of points of all loops.

        Returns:
            list: List containing the number of points of all loops.
        """
        shape = []
        loop: Loop | None = self
        while loop is not None:
            shape.append(int(loop.num))
            loop = loop.loop
        return shape

    @property
    def num_loops(self) -> int:
        """Loop 'num_loops' property.

        Returns:
            int: Number of nested loops.
        """
        return len(self.loops)

    @property
    def loops(self) -> list[Loop]:
        """Loop 'loops' property.

        Returns:
            list[Loop]: List of loop objects.
        """
        loops = []
        loop: Loop | None = self
        while loop is not None:
            loops.append(loop)
            loop = loop.loop
        return loops

    def to_dict(self) -> dict:
        """Convert class to a dictionary.

        Returns:
            dict: Dictionary representation of the class.
        """
        return {
            LOOP.ALIAS: self.alias,
            LOOP.PARAMETER: self.parameter.value,
            LOOP.VALUES: self.values.tolist(),
            LOOP.LOOP: self.loop.to_dict() if self.loop is not None else None,
            LOOP.CHANNEL_ID: self.channel_id,
        }

    @property
    def start(self):
        """returns 'start' options property."""
        return self.values[0]

    @property
    def stop(self):
        """returns 'stop' options property."""
        return self.values[-1]

    @property
    def num(self):
        """returns 'num' options property."""
        return len(self.values)

    def __eq__(self, other: object) -> bool:
        """Equality operator"""
        if not isinstance(other, Loop):
            return False
        return (
            self.alias == other.alias
            and self.parameter == other.parameter
            and self.loop == other.loop
            and self.channel_id == other.channel_id
            and (self.values == other.values).all()
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access
import math
from collections import deque
from typing import Any, Callable

import numpy as np
import qpysequence as QPy
import qpysequence.program as QPyProgram
import qpysequence.program.instructions as QPyInstructions
from qpysequence.utils.constants import INST_MAX_WAIT

from qililab.qprogram.blocks import Average, Block, ForLoop, InfiniteLoop, Loop, Parallel
from qililab.qprogram.operations import (
    Acquire,
    Operation,
    Play,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.qprogram import QProgram
from qililab.qprogram.variable import Variable
from qililab.waveforms import IQPair, Waveform


class BusCompilationInfo:  # pylint: disable=too-many-instance-attributes, too-few-public-methods
    """Class representing the information stored by QBloxCompiler for a bus."""

    def __init__(self) -> None:
        # The generated Sequence
        self.qpy_sequence = QPy.Sequence(
            program=QPy.Program(), waveforms=QPy.Waveforms(), acquisitions=QPy.Acquisitions(), weights=QPy.Weights()
        )

        # Dictionaries to hold mappings useful during compilation.
        self.variable_to_register: dict[Variable, QPyProgram.Register] = {}
        self.waveform_to_index: dict[str, int] = {}
        self.weight_to_index: dict[str, int] = {}
        self.acquisition_to_index: dict[str, int] = {}

        # Create and append the main block to the Sequence's program
        main_block = QPyProgram.Block(name="main")
        self.qpy_sequence._program.append_block(main_block)

        # Stacks to manage block hierarchy during compilation
        self.qpy_block_stack: deque[QPyProgram.Block] = deque([main_block])
        self.qprogram_block_stack: deque[Block] = deque()

        # Counters to help with naming and indexing
        self.next_bin_index = 0
        self.next_acquisition_index = 0
        self.loop_counter = 0
        self.average_counter = 0

        # Syncing durations
        self.static_duration = 0
        self.dynamic_durations: list[Variable] = []
        self.sync_durations: list[QPyProgram.Register] = []

        # Syncing marker. If true, a real-time instruction has been added since the last sync or the beginning of the program.
        self.marked_for_sync = False


class QbloxCompiler:  # pylint: disable=too-few-public-methods
    """A class for compiling QProgram to QBlox hardware."""

    minimum_wait_duration: int = 4

    def __init__(self) -> None:
        # Handlers to map each operation to a corresponding handler function
        self._handlers: dict[type, Callable] = {
            InfiniteLoop: self._handle_infinite_loop,
            Parallel: self._handle_parallel,
            Average: self._handle_average,
            ForLoop: self._handle_for_loop,
            Loop: self._handle_loop,
            SetFrequency: self._handle_set_frequency,
            SetPhase: self._handle_set_phase,
            ResetPhase: self._handle_reset_phase,
            SetGain: self._handle_set_gain,
            SetOffset: self._handle_set_offset,
            Wait: self._handle_wait,
            Sync: self._handle_sync,
            Acquire: self._handle_acquire,
            Play: self._handle_play,
        }

        self._qprogram: QProgram
        self._bus_mapping: dict[str, str] | None
        self._buses: dict[str, BusCompilationInfo]
        self._sync_counter: int

    def compile(self, qprogram: QProgram, bus_mapping: dict[str, str] | None = None) -> dict[str, QPy.Sequence]:
        """Compile QProgram to qpysequence.Sequence

        Args:
            qprogram (QProgram): The QProgram to be compiled
            bus_mapping (dict[str, str] | None, optional): Optional mapping of bus names. Defaults to None.

        Returns:
            dict[str, QPy.Sequence]: A dictionary with the buses participating in the QProgram as keys and the corresponding Sequence as values.
        """

        def traverse(block: Block):
            for bus in self._buses:
                self._buses[bus].qprogram_block_stack.append(block)
            for element in block.elements:
                handler = self._handlers.get(type(element))
                if not handler:
                    raise NotImplementedError(f"{element.__class__} is currently not supported in QBlox.")
                appended = handler(element)
                if isinstance(element, Block):
                    traverse(element)
                    if not self._qprogram.disable_autosync and isinstance(element, (ForLoop, Parallel, Loop, Average)):
                        self._handle_sync(element=Sync(buses=None))
                    if appended:
                        for bus in self._buses:
                            self._buses[bus].qpy_block_stack.pop()
            for bus in self._buses:
                self._buses[bus].qprogram_block_stack.pop()

        self._qprogram = qprogram
        self._bus_mapping = bus_mapping
        self._sync_counter = 0
        self._buses = self._populate_buses()

        # Recursive traversal to convert QProgram blocks to Sequence
        traverse(self._qprogram._body)

        # Post-processing: Add stop instructions and compile
        for bus in self._buses:
            self._buses[bus].qpy_block_stack[0].append_component(component=QPyInstructions.Stop())
            self._buses[bus].qpy_sequence._program.compile()

        # Return a dictionary with bus names as keys and the compiled Sequence as values.
        return {
            self._bus_mapping[bus] if self._bus_mapping and bus in self._bus_mapping else bus: bus_info.qpy_sequence
            for bus, bus_info in self._buses.items()
        }

    def _populate_buses(self):
        """Map each bus in the QProgram to a BusCompilationInfo instance.

        Returns:
            A dictionary where the keys are bus names and the values are BusCompilationInfo objects.
        """

        return {bus: BusCompilationInfo() for bus in self._qprogram.buses}

    def _append_to_waveforms_of_bus(self, bus: str, waveform_I: Waveform, waveform_Q: Waveform | None):
        """Append waveforms to Sequence's Waveforms of the given bus.

        Args:
            bus (str): Name of the bus.
            waveform_I (Waveform): I waveform.
            waveform_Q (Waveform | None): Q waveform.
        """

        def handle_waveform(waveform: Waveform | None, default_length: int = 0):
            _hash = QbloxCompiler._hash_waveform(waveform) if waveform else f"zeros {default_length}"

            if _hash in self._buses[bus].waveform_to_index:
                index = self._buses[bus].waveform_to_index[_hash]
                length = next(
                    len(waveform.data)
                    for waveform in self._buses[bus].qpy_sequence._waveforms._waveforms
                    if waveform.index == index
                )
                return index, length

            envelope = waveform.envelope() if waveform else np.zeros(default_length)
            index = self._buses[bus].qpy_sequence._waveforms.add(envelope)
            self._buses[bus].waveform_to_index[_hash] = index
            return index, len(envelope)

        index_I, length_I = handle_waveform(waveform_I, 0)
        index_Q, _ = handle_waveform(waveform_Q, len(waveform_I.envelope()))
        return index_I, index_Q, length_I

    def _append_to_weights_of_bus(self, bus: str, weights: IQPair):
        def handle_waveform(waveform: Waveform):
            _hash = QbloxCompiler._hash_waveform(waveform)

            if _hash in self._buses[bus].weight_to_index:
                index = self._buses[bus].weight_to_index[_hash]
                length = next(
                    len(weight.data)
                    for weight in self._buses[bus].qpy_sequence._weights._weights  # pylint: disable=protected-access
                    if weight.index == index
                )
                return index, length

            envelope = waveform.envelope()
            length = len(envelope)
            index = self._buses[bus].qpy_sequence._weights.add(envelope)  # pylint: disable=protected-access
            self._buses[bus].weight_to_index[_hash] = index
            return index, length

        index_I, length_I = handle_waveform(weights.I)
        index_Q, _ = handle_waveform(weights.Q)
        return index_I, index_Q, length_I

    def _handle_parallel(self, element: Parallel):
        if not element.loops:
            raise NotImplementedError("Parallel block should contain loops.")
        if any(isinstance(loop, Loop) for loop in element.loops):
            raise NotImplementedError("Loops with arbitrary numpy arrays are not currently supported for QBlox.")

        loops = []
        iterations = []
        for loop in element.loops:
            operation = QbloxCompiler._get_reference_operation_of_loop(loop=loop, starting_block=element)
            start, step, iters = QbloxCompiler._convert_for_loop_values(loop, operation)  # type: ignore[arg-type]
            loops.append((start, step))
            iterations.append(iters)
        iterations = min(iterations)

        # iterations = min(QbloxCompiler._calculate_iterations(loop.start, loop.stop, loop.step) for loop in element.loops)
        # loops = [(QbloxCompiler._convert_for_loop_values(for_loop=loop, operation=QbloxCompiler._get_reference_operation_of_loop(element)))[:2]) for loop in element.loops]

        for bus in self._buses:
            qpy_loop = QPyProgram.IterativeLoop(
                name=f"loop_{self._buses[bus].loop_counter}", iterations=iterations, loops=loops
            )
            for i, loop in enumerate(element.loops):
                self._buses[bus].variable_to_register[loop.variable] = qpy_loop.loop_registers[i]
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].loop_counter += 1
        return True

    def _handle_average(self, element: Average):
        for bus in self._buses:
            qpy_loop = QPyProgram.Loop(name=f"avg_{self._buses[bus].average_counter}", begin=element.shots)
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].average_counter += 1
        return True

    def _handle_infinite_loop(self, _: InfiniteLoop):
        for bus in self._buses:
            qpy_loop = QPyProgram.InfiniteLoop(name=f"infinite_loop_{self._buses[bus].loop_counter}")
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].loop_counter += 1
        return True

    def _handle_for_loop(self, element: ForLoop):
        operation = QbloxCompiler._get_reference_operation_of_loop(element)
        start, step, iterations = QbloxCompiler._convert_for_loop_values(element, operation)
        for bus in self._buses:
            qpy_loop = QPyProgram.IterativeLoop(
                name=f"loop_{self._buses[bus].loop_counter}", iterations=iterations, loops=[(start, step)]
            )
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].variable_to_register[element.variable] = qpy_loop.loop_registers[0]
            self._buses[bus].loop_counter += 1
        return True

    def _handle_loop(self, _: Loop):
        raise NotImplementedError("Loops with arbitrary numpy arrays are not currently supported for QBlox.")

    def _handle_set_frequency(self, element: SetFrequency):
        convert = QbloxCompiler._convert_value(element)
        frequency = (
            self._buses[element.bus].variable_to_register[element.frequency]
            if isinstance(element.frequency, Variable)
            else convert(element.frequency)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetFreq(frequency=frequency)
        )

    def _handle_set_phase(self, element: SetPhase):
        convert = QbloxCompiler._convert_value(element)
        phase = (
            self._buses[element.bus].variable_to_register[element.phase]
            if isinstance(element.phase, Variable)
            else convert(element.phase)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(component=QPyInstructions.SetPh(phase=phase))

    def _handle_reset_phase(self, element: ResetPhase):
        self._buses[element.bus].qpy_block_stack[-1].append_component(component=QPyInstructions.ResetPh())

    def _handle_set_gain(self, element: SetGain):
        convert = QbloxCompiler._convert_value(element)
        gain = (
            self._buses[element.bus].variable_to_register[element.gain]
            if isinstance(element.gain, Variable)
            else convert(element.gain)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetAwgGain(gain_0=gain, gain_1=gain)
        )

    def _handle_set_offset(self, element: SetOffset):
        convert = QbloxCompiler._convert_value(element)
        offset_0 = (
            self._buses[element.bus].variable_to_register[element.offset_path0]
            if isinstance(element.offset_path0, Variable)
            else convert(element.offset_path0)
        )
        offset_1 = (
            self._buses[element.bus].variable_to_register[element.offset_path1]
            if isinstance(element.offset_path1, Variable)
            else convert(element.offset_path1)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetAwgOffs(offset_0=offset_0, offset_1=offset_1)
        )

    def _handle_wait(self, element: Wait):
        duration: QPyProgram.Register | int
        if isinstance(element.duration, Variable):
            duration = self._buses[element.bus].variable_to_register[element.duration]
            self._buses[element.bus].dynamic_durations.append(element.duration)
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Wait(wait_time=duration)
            )
        else:
            convert = QbloxCompiler._convert_value(element)
            duration = convert(element.duration)
            self._buses[element.bus].static_duration += duration
            # loop over wait instructions if static duration is longer than allowed qblox max wait time of 2**16 -4
            self._handle_add_waits(bus=element.bus, duration=duration)

        self._buses[element.bus].marked_for_sync = True

    def _handle_add_waits(self, bus: str, duration: int):
        """Wait for longer than QBLOX INST_MAX_WAIT by looping over wait instructions

        Args:
            element (Wait): wait element
            duration (int): duration to wait in ns
        """
        if duration > INST_MAX_WAIT:
            for _ in range(duration // INST_MAX_WAIT):
                self._buses[bus].qpy_block_stack[-1].append_component(
                    component=QPyInstructions.Wait(wait_time=INST_MAX_WAIT)
                )
        # add the remaining wait time (or all of it if the above conditional is false)
        self._buses[bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.Wait(wait_time=duration % INST_MAX_WAIT)
        )

    def _handle_sync(self, element: Sync):
        # Get the buses involved in the sync operation.
        buses = set(element.buses or self._buses)

        # If they are zero or one, return.
        if len(buses) <= 1:
            return

        # If there is no bus marked for sync, return.
        if all(not self._buses[bus].marked_for_sync for bus in buses):
            return

        # Is there any bus that has dynamic durations?
        if any(bus for bus in buses if self._buses[bus].dynamic_durations or self._buses[bus].sync_durations):
            # If yes, we must add a sync block that calculates the difference between buses dynamically.
            # But the following doesn't work unfortunetely, so raise an error for now.
            self.__handle_dynamic_sync(buses=buses)
        else:
            # If no, calculating the difference is trivial.
            self.__handle_static_sync(buses=buses)

        # In any case, mark al buses as synced.
        for bus in buses:
            self._buses[bus].marked_for_sync = False

    def __handle_static_sync(self, buses: set[str]):
        max_duration = max(self._buses[bus].static_duration for bus in buses)
        for bus in buses:
            duration_diff = max_duration - self._buses[bus].static_duration
            if duration_diff > 0:
                # loop over wait instructions if static duration is longer than allowed qblox max wait time of 2**16 -4
                self._handle_add_waits(bus=bus, duration=duration_diff)
                self._buses[bus].static_duration += duration_diff

    def __handle_dynamic_sync(self, buses: set[str]):
        raise NotImplementedError("Dynamic syncing is not implemented yet.")

    def _handle_acquire(self, element: Acquire):
        loops = [
            (i, loop)
            for i, loop in enumerate(self._buses[element.bus].qpy_block_stack)
            if isinstance(loop, QPyProgram.IterativeLoop) and not loop.name.startswith("avg")
        ]
        num_bins = math.prod(loop[1].iterations for loop in loops)
        self._buses[element.bus].qpy_sequence._acquisitions.add(
            name=f"acquisition_{self._buses[element.bus].next_acquisition_index}",
            num_bins=num_bins,
            index=self._buses[element.bus].next_acquisition_index,
        )

        index_I, index_Q, integration_length = self._append_to_weights_of_bus(element.bus, weights=element.weights)

        if num_bins == 1:
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.AcquireWeighed(
                    acq_index=self._buses[element.bus].next_acquisition_index,
                    bin_index=self._buses[element.bus].next_bin_index,
                    weight_index_0=index_I,
                    weight_index_1=index_Q,
                    wait_time=integration_length,
                )
            )
        else:
            bin_register = QPyProgram.Register()
            block_index_for_move_instruction = loops[0][0] - 1 if loops else -2
            block_index_for_add_instruction = loops[-1][0] if loops else -1
            self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                component=QPyInstructions.Move(var=self._buses[element.bus].next_bin_index, register=bin_register),
                bot_position=len(self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components),
            )
            register_I, register_Q = QPyProgram.Register(), QPyProgram.Register()
            self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                component=QPyInstructions.Move(var=index_I, register=register_I),
                bot_position=len(self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components),
            )
            self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                component=QPyInstructions.Move(var=index_Q, register=register_Q),
                bot_position=len(self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components),
            )
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.AcquireWeighed(
                    acq_index=self._buses[element.bus].next_acquisition_index,
                    bin_index=bin_register,
                    weight_index_0=register_I,
                    weight_index_1=register_Q,
                    wait_time=integration_length,
                )
            )
            self._buses[element.bus].qpy_block_stack[block_index_for_add_instruction].append_component(
                component=QPyInstructions.Add(origin=bin_register, var=1, destination=bin_register)
            )
        self._buses[element.bus].static_duration += integration_length
        self._buses[element.bus].next_bin_index = 0  # maybe this counter can be removed completely
        self._buses[element.bus].next_acquisition_index += 1
        self._buses[element.bus].marked_for_sync = True

    def _handle_play(self, element: Play):
        waveform_I, waveform_Q = element.get_waveforms()
        waveform_variables = element.get_waveform_variables()
        if not waveform_variables:
            index_I, index_Q, duration = self._append_to_waveforms_of_bus(
                bus=element.bus, waveform_I=waveform_I, waveform_Q=waveform_Q
            )
            if element.wait_time is not None:  # TODO: Change this in clean fix
                duration = element.wait_time
            convert = QbloxCompiler._convert_value(element)
            duration = convert(duration)
            self._buses[element.bus].static_duration += duration
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Play(index_I, index_Q, wait_time=duration)
            )
            self._buses[element.bus].marked_for_sync = True

    @staticmethod
    def _get_reference_operation_of_loop(loop: Loop | ForLoop, starting_block: Block | None = None):
        def collect_operations(block: Block):
            for element in block.elements:
                if isinstance(element, Block):
                    yield from collect_operations(element)
                else:
                    if any(variable == loop.variable for variable in element.get_variables()):
                        yield element

        starting_block = starting_block or loop
        operations = list(collect_operations(starting_block))

        if not operations:
            return None
        if isinstance(operations[0], Play) and operations[0].get_waveform_variables():
            raise NotImplementedError("TODO: Variables referenced in a loop cannot be used in Play operation.")
        return operations[0]

    @staticmethod
    def _calculate_iterations(start: int | float, stop: int | float, step: int | float):
        if step == 0:
            raise ValueError("Step value cannot be zero")

        # Calculate the raw number of iterations
        raw_iterations = (stop - start + step) / step

        # If the raw number of iterations is very close to an integer, round it to that integer
        # This accounts for potential floating-point inaccuracies
        if abs(raw_iterations - round(raw_iterations)) < 1e-9:
            return round(raw_iterations)

        # Otherwise, if we're incrementing, take the ceiling, and if we're decrementing, take the floor
        return math.floor(raw_iterations) if step > 0 else math.ceil(raw_iterations)

    @staticmethod
    def _convert_for_loop_values(for_loop: ForLoop, operation: Operation):
        convert = QbloxCompiler._convert_value(operation)
        iterations = QbloxCompiler._calculate_iterations(start=for_loop.start, stop=for_loop.stop, step=for_loop.step)
        qblox_start = convert(for_loop.start)
        qblox_stop = convert(for_loop.stop)
        qblox_step = (qblox_stop - qblox_start) // (iterations - 1)
        return (qblox_start, qblox_step, iterations)

    @staticmethod
    def _convert_value(operation: Operation) -> Callable[[Any], int]:
        conversion_map: dict[type[Operation], Callable[[Any], int]] = {
            SetFrequency: lambda x: int(x * 4),
            SetPhase: lambda x: int(x * 1e9 / (2 * np.pi)),
            SetGain: lambda x: int(x * 32_767),
            SetOffset: lambda x: int(x * 32_767),
            Wait: lambda x: int(max(x, QbloxCompiler.minimum_wait_duration)),
            Play: lambda x: int(max(x, QbloxCompiler.minimum_wait_duration)),
        }
        return conversion_map.get(type(operation), lambda x: int(x))  # pylint: disable=unnecessary-lambda

    @staticmethod
    def _hash_waveform(waveform: Waveform):
        hashes = {
            key: (value.__dict__ if isinstance(value, Waveform) else value) for key, value in waveform.__dict__.items()
        }
        return f"{waveform.__class__.__name__} {hashes}"
